# IssueTrackerProduct
#
# Peter Bengtsson <mail@peterbe.com>
# License: ZPL
#

__doc__="""IssueTrackerProduct is the easiest bug/issue tracker
system to use for Zope.
By Peter Bengtsson <mail@peterbe.com>

Credits:
Jo Meder for addhrefs() function
Jeff Robson of Cynergic Net Solutions for the encoding of email hyperlinks idea
"""
__version__='0.5.0b'

# python
import string
import os
import re
import sys
import random
import copy
import poplib
from types import IntType, StringType, ListType, TupleType
from math import floor

try:
    import email.Parser as email_Parser
except ImportError:
    email_Parser = None

try:
    from stripogram import html2text, html2safehtml
except ImportError:
    html2text = html2safehtml = None
    
    
    
    
# Zope
from Products.PageTemplates.PageTemplateFile import PageTemplateFile as PTF
from Globals import Persistent, InitializeClass, package_home, DTMLFile
from Products.PythonScripts.standard import html_quote, url_quote_plus
from Products.PythonScripts.PythonScript import PythonScript
from OFS import SimpleItem, ObjectManager, Folder
from DocumentTemplate import sequence
from AccessControl import ClassSecurityInfo
from Products.ZCatalog.CatalogAwareness import CatalogAware
from zLOG import LOG, ERROR, INFO, PROBLEM, WARNING
from DateTime import DateTime

# Product
import IssueTrackerNotifyables
import IssueTrackerEmail
import IssueTrackerUtils
from IssueTrackerUtils import cookIdAndTitle, ss, getNowdate,  \
                              fixDictofLists, getFromstring
from IssueTrackerConstants import *


manage_addIssueTrackerForm = PTF('zpt/addIssueTrackerForm', globals())

def manage_addIssueTracker(dispatcher, id, title='', REQUEST=None):
    """ add IssueTracker instance via the web """
    dest = dispatcher.Destination()
    issuetracker = IssueTracker(id, title)
    dest._setObject(id, issuetracker)
    self = dest._getOb(id)
    
    
    self.DeployStandards()
    self.InitZCatalog()

    if REQUEST is not None:
        # whereto next?
        if REQUEST.has_key('addandedit'):
            REQUEST.RESPONSE.redirect(self.absolute_url()+'/PropertiesWizard')
        elif REQUEST.has_key('DestinationURL'):
            REQUEST.RESPONSE.redirect(REQUEST.DestinationURL+'/manage_workspace')
        else:
            REQUEST.RESPONSE.redirect(REQUEST.URL1+'/manage_workspace')
    
    

class IssueTracker(Folder.Folder, CatalogAware, Persistent,
                   IssueTrackerNotifyables.Notifyables
                   ):
    """ IssueTracker class """

    meta_type = ISSUETRACKER_METATYPE

    security=ClassSecurityInfo()

    manage_options = Folder.Folder.manage_options[:2] + \
        ({'label':'Properties', 'action':'manage_editIssueTrackerPropertiesForm'},
         {'label':'Management', 'action':'manage_ManagementForm'}, \
         {'label':'POP3',       'action':'manage_POP3ManagementForm'}) \
        + Folder.Folder.manage_options[3:]


    
    def __init__(self, id, title='',
                 sitemaster_name=DEFAULT_SITEMASTER_NAME,
                 sitemaster_email=DEFAULT_SITEMASTER_EMAIL):
        """ Init IssueTracker class """
        self.id = str(id)
        self.title = str(title)
        
        self.types              = DEFAULT_TYPES
        self.urgencies          = DEFAULT_URGENCIES
        self.sections_options   = DEFAULT_SECTIONS_OPTIONS
        self.defaultsections    = DEFAULT_SECTIONS
        self.when_ignore_word   = DEFAULT_WHEN_IGNORE_WORD
        self.display_date       = DEFAULT_DISPLAY_DATE
        self.always_notify      = DEFAULT_ALWAYS_NOTIFY
        self.sitemaster_name    = sitemaster_name
        self.sitemaster_email   = sitemaster_email
        self.odd_bgcolor        = DEFAULT_ODD_BGCOLOR
        self.even_bgcolor       = DEFAULT_EVEN_BGCOLOR
        self.default_type       = DEFAULT_TYPE
        self.default_urgency    = DEFAULT_URGENCY
        self.manager_roles      = DEFAULT_MANAGER_ROLES
        self.default_batch_size = DEFAULT_DEFAULT_BATCH_SIZE
        self.issueprefix        = DEFAULT_ISSUEPREFIX
        self.no_fileattachments = DEFAULT_NO_FILEATTACHMENTS
        self.no_followup_fileattachments = DEFAULT_NO_FOLLOWUP_FILEATTACHMENTS
        self.statuses           = DEFAULT_STATUSES
        self.statuses_verbs     = DEFAULT_STATUSES_VERBS
        self.display_formats    = DEFAULT_DISPLAY_FORMATS
        self.default_display_format = DEFAULT_DEFAULT_DISPLAY_FORMAT
        self.dispatch_on_submit = DEFAULT_DISPATCH_ON_SUBMIT
        self.randomid_length = DEFAULT_RANDOMID_LENGTH
        self.allow_issueattrchange = DEFAULT_ALLOW_ISSUEATTRCHANGE
        self.stop_cache         = DEFAULT_STOP_CACHE
        self.allow_subscription = DEFAULT_ALLOW_SUBSCRIPTION
        self.private_statistics = DEFAULT_PRIVATE_STATISTICS
        self.show_confidential_option = DEFAULT_SHOW_CONFIDENTIAL_OPTION
        self.show_hideme_option = DEFAULT_SHOW_HIDEME_OPTION
        self.show_download_button = DEFAULT_SHOW_DOWNLOAD_BUTTON
        self.encode_emaildisplay = DEFAULT_ENCODE_EMAILDISPLAY

    # forms
    manage_editIssueTrackerPropertiesForm = \
                    DTMLFile('dtml/editIssueTrackerPropertiesForm',\
                             globals(), native_properties=NATIVE_PROPERTIES)
    manage_ManagementForm = DTMLFile('dtml/ManagementForm', globals())
    manage_POP3ManagementForm = DTMLFile('dtml/POP3ManagementForm', globals())
    security.declareProtected('View management screens', 'PropertiesWizard')    
    PropertiesWizard = DTMLFile('dtml/PropertiesWizard', globals())

    # scripts

    def timeSince(self, firstdate, seconddate, afterword=None):
        """ wrap script """
        script = IssueTrackerUtils.timeSince
        return script(firstdate, seconddate, afterword)
    
    def ShowDescription(self, text, display_format=''):
        """ pass on to utilities module """
        script = IssueTrackerUtils.ShowDescription
        return script(text, display_format)

    def ShowFilesize(self, bytes):
        """ pass on to utilities module """
        return IssueTrackerUtils.ShowFilesize(bytes)

    def ValidEmailAddress(self, address):
        """ pass on to utilities module """
        return IssueTrackerUtils.ValidEmailAddress(address)
    
    ## This is "under construction"
    #def standard_error_message(self, REQUEST, RESPONSE=None):
        #""" custom error message """
        #return getattr(self, 'standard_error_message.html')

    def getTitle(self):
        """ return title """
        return self.title

    def getStatusesVerbs(self):
        """ return statuses_verbs """
        return getattr(self, 'statuses_verbs', DEFAULT_STATUSES_VERBS)
    
    def getStatuses(self):
        """ return statuses """
        return self.statuses
    
    def getStatusesMerged(self, aslist=0, asdict=0):
        """ return statuses and statuses_verbs next to each other 
        So it looks like this ['taken, take', 'rejected, reject', ...]"""
        statuses = self.getStatuses()
        verbs = self.getStatusesVerbs()
        nl=[]
        nldict = {}
        delimiter = ', '
        for i in range(len(statuses)):
            nldict[statuses[i].strip()] = verbs[i].strip()
            if aslist:
                nl.append([statuses[i], verbs[i]])
            else:
                nl.append(statuses[i]+delimiter+verbs[i])
        if asdict:
            return nldict
        else:
            return nl
        
    def splitStatusesAndVerbs(self, statuses_and_verbs):
        """ list might be ['open, open', 'taken, take', ...]
        then split this up into two lists """
        statuses = []
        verbs = []
        for each in statuses_and_verbs:
            found_delim = max(each.find(','), each.find(';'),
                              each.find('|'))
            if found_delim > -1:
                splitted = [each[:found_delim], each[found_delim+1:]]
                statuses.append(splitted[0].strip())
                if splitted[1].strip() == '':
                    verbs.append(splitted[0].strip())
                else:
                    verbs.append(splitted[1].strip())
            else:
                statuses.append(each)
                verbs.append(each)
                
        return statuses, verbs

    
    def getSectionOptions(self):
        """ return section options """
        return self.sections_options
        

    def manage_editIssueTrackerProperties(self, REQUEST=None):
        """ save all IssueTracker related issues """
        hk = self.REQUEST.has_key
        get = self.REQUEST.get
        strings = ['title','when_ignore_word','display_date','sitemaster_name',
                   'sitemaster_email','odd_bgcolor','even_bgcolor',
                   'default_type','default_urgency','issueprefix',
                   'default_display_format']
        lists = ['types','urgencies','sections_options','defaultsections',
                 'statuses','statuses_verbs','display_formats',
                 'manager_roles','always_notify']
        ints = ['default_batch_size','randomid_length','no_fileattachments',
                'no_followup_fileattachments']
        booleans = ['dispatch_on_submit','allow_issueattrchange','stop_cache',
                    'allow_subscription','private_statistics',
                    'show_confidential_option','show_hideme_option',
                    'show_download_button','encode_emaildisplay']
        dict = self.__dict__
        for each in strings:
            if hk(each) and type(get(each))==StringType:
                dict[each] = get(each).strip()

        for each in ints:
            if hk(each):
                if type(get(each))==IntType:
                    dict[each] = get(each)
                else:
                    LOG(self.meta_type, WARNING,
                        '%s not integer'%get(each), '')
        for each in lists:
            if hk(each) and type(get(each))==ListType:
                dict[each] = get(each)

        for each in booleans:
            if hk(each) and get(each):
                dict[each] = true
            else:
                dict[each] = false
                
        # now for a special one
        if hk('statuses-and-verbs'):
            if type(get('statuses-and-verbs'))==ListType:
                L1, L2 = self.splitStatusesAndVerbs(get('statuses-and-verbs'))
                self.statuses = L1
                self.statuses_verbs = L2
            else:
                LOG(self.__class__.__name__, PROBLEM,
                    "Statuses and verbs not list type")
        # for the custom properties
        if REQUEST is not None:
            self.manage_editProperties(REQUEST)
        return self.manage_editIssueTrackerPropertiesForm(self.REQUEST,
                        manage_tabs_message='IssueTracker properties updated.')


    security.declareProtected('View management screens','manage_addOtherProperty')
    def manage_addOtherProperty(self, id, value, type):
        """ Add arbitrary property """
        self.manage_addProperty(id, value, type)
        page = self.manage_editIssueTrackerPropertiesForm
        return page(self.REQUEST, manage_tabs_message='Other property added.')

    security.declareProtected('View management screens','manage_delOtherProperties')
    def manage_delOtherProperties(self, ids):
        """ remove arbitrary properties """
        self.manage_delProperties(ids)
        page = self.manage_editIssueTrackerPropertiesForm
        return page(self.REQUEST)
    
    def ReplaceEmail(self, old, new, caseinsensitive=1, REQUEST=None):
        """ Method that lets you change an occurance of an email address
            to another.
            Useful if a frequence user has changed email accout or something.
        """
        if caseinsensitive:
            old = old.lower()
        root = self.getRoot()
        nochanges_issues = 0
        nochanges_threads = 0
        for issue in root.objectValues(ISSUE_METATYPE):
            iemail = issue.email
            if caseinsensitive:
                iemail = iemail.lower()
            if iemail == old:
                issue.email = new
                nochanges_issues = nochanges_issues + 1
            for thread in issue.objectValues(ISSUETHREAD_METATYPE):
                temail = thread.email
                if caseinsensitive:
                    temail = temail.lower()
                if temail == old:
                    thread.email = new
                    nochanges_threads = nochanges_threads + 1
        
        msg = "Changed %s issues and %s threads"%\
              (nochanges_issues, nochanges_threads)
        if REQUEST is None:
            return msg
        else:
            method = IssueTrackerUtils.AddParam2URL
            desturl = root.absolute_url()+"/manage_ManagementForm"
            url = method(desturl,{'manage_tabs_message':msg})
            self.REQUEST.RESPONSE.redirect(url)
            
        
    def AssertAllProperties(self, REQUEST=None):
        """ invoke the assertAllProperties() on all objects """
        count = 0
        root = self.getRoot()
        for issue in root.objectValues(ISSUE_METATYPE):
            count += issue.assertAllProperties()
            
        msg = "Made sure all objects have all properties"
        if REQUEST is None:
            return msg
        else:
            root = self.getRoot()
            method = IssueTrackerUtils.AddParam2URL
            desturl = root.absolute_url()+"/manage_ManagementForm"
            url = method(desturl,{'manage_tabs_message':msg})
            self.REQUEST.RESPONSE.redirect(url)
            

    def DeployStandards(self, remove_oldstuff=0, DestinationURL=None,
                        initzcatalog=1):
        """ copy images and other documents into the instance unless they
            are already there
        """
        t={}
        
        if initzcatalog:
            self.InitZCatalog()
            
            
        # create folders
        root = self.getRoot()
        for f in ['notifyables', 'www','pop3']:
            if not hasattr(root, f):
                root.manage_addFolder(f)
                t[f]='Folder'

        osj = os.path.join
        standards_home = osj(package_home(globals()),'standards')
        t = self._deployImages(root, standards_home,
                               t=t, remove_oldstuff=remove_oldstuff)

        www_home = osj(standards_home,'www')
        t = self._deployImages(root.www, www_home,
                               t=t, remove_oldstuff=remove_oldstuff)
        
        # shortcut
        addPG = root.manage_addProduct['PageTemplates'].manage_addPageTemplate
        
        for filestr in os.listdir(standards_home):
            if filestr[-5:] == '.dtml':
                id, title = cookIdAndTitle(filestr.replace('.dtml',''))
                if hasattr(root, id) and remove_oldstuff:
                    root.manage_delObjects([id])
                
                if not hasattr(root, id):
                    file = DTMLFile('standards/%s'%filestr.replace('.dtml',''),
                                    globals()).read()
                    root.manage_addDTMLDocument(id, title, file=file)
                    t[id] ="DTML Document"
            elif filestr[-4:] == '.zpt':
                id, title = cookIdAndTitle(filestr.replace('.zpt',''))
                if hasattr(root, id) and remove_oldstuff:
                    root.manage_delObjects([id])
                    
                if not hasattr(root, id):
                    file = open(osj(standards_home,filestr)).read()
                    addPG(id, title=title, text=file)
                    t[id]="Page Template"                    
            elif filestr[-3:] == '.py':
                id, title = cookIdAndTitle(filestr.replace('.py',''))
                if hasattr(root, id) and remove_oldstuff:
                    root.manage_delObjects([id])
                    
                if not hasattr(root, id):
                    file = open(osj(standards_home, filestr)).read()
                    id = root._setObject(id, PythonScript(id))
                    root._getOb(id).write(file)
                    t[id]="Script (Python)"
        if DestinationURL:
            msg = ''
            for k,v in t.items():
                msg = "%s (%s)\n%s"%(k,v,msg)
            method = IssueTrackerUtils.AddParam2URL
            url = method(DestinationURL,{'manage_tabs_message':\
                                "Standard objects deployed\n\n%s"%msg})
            self.REQUEST.RESPONSE.redirect(url)
        else:
            return "Standard objects deployed\n%s"%t

    def _deployImages(self, destination, dir, extensions=['.gif','.ico'], t={},
                      remove_oldstuff=0):
        """ do the actual deployment of images in a dir """
        osj = os.path.join
        for filestr in os.listdir(dir):
            if self._file_has_extensions(filestr, extensions):
                # take the image
                id, title = cookIdAndTitle(filestr)
                base= getattr(destination,'aq_base',destination)
                if hasattr(base, id) and remove_oldstuff:
                    destination.manage_delObjects([id])
                    
                if not hasattr(base, id):
                    destination.manage_addImage(id, title=title, \
                          file=open(osj(dir, filestr),'rb').read())
                    t[id]="Image"
        return t
    
    def _file_has_extensions(self, filestr, extensions):
        """ check if a filestr has any of the give extensions """
        for extension in extensions:
            if filestr.find(extension) > -1:
                return true
        return false

                
    def InitZCatalog(self):
        """ create a ZCatalog called 'ICatalog' and change its properties
        accordingly """
        if not 'ICatalog' in self.objectIds('ZCatalog'):
            self.manage_addProduct['ZCatalog'].manage_addZCatalog('ICatalog','')
        zcatalog = self.getCatalog()
        indexes = zcatalog._catalog.indexes
        
        if not indexes.has_key('id'):
            zcatalog.addIndex('id', 'FieldIndex')
        add=('title','status','description','comment','fromname',
             'email','url2issue')
        for adder in add:
            if not indexes.has_key(adder):
                zcatalog.addIndex(adder, 'TextIndex')


    def getRoot(self):
        """ Get the root instance object """
        mtype = ISSUETRACKER_METATYPE
        if self.meta_type == mtype:
            r = self
        elif self.aq_parent.meta_type == mtype:
            r = self.aq_parent
        elif self.aq_parent.aq_parent.meta_type == mtype:
            r = self.aq_parent.aq_parent
        elif self.aq_parent.aq_parent.aq_parent.meta_type == mtype:
            r = self.aq_parent.aq_parent.aq_parent
        else:
            r = None
        
        return r

    def getRootURL(self):
        """ quick wrapper around getRoot() """
        return self.getRoot().absolute_url()

    def getCookiekey(self, which):
        """ return the cookiekey constants depending on key """
        which = which.lower().strip().replace('_','')
        if which=='name':
            return NAME_COOKIEKEY
        elif which=='email':
            return EMAIL_COOKIEKEY
        elif which=='displayformat':
            return DISPLAY_FORMAT_COOKIEKEY
        elif which=='sortorder':
            return SORTORDER_COOKIEKEY
        elif which=='sortorderreverse':
            return SORTORDER_REVERSE_COOKIEKEY
        
    def __before_publishing_traverse__(self, object, request):
        """ sort things out before publising object """
        self.get_environ()

    def get_environ(self):
        """ Populate REQUEST as appropriate """
        request = self.REQUEST
        stack = request['TraversalRequestNameStack']
        popped = []

        _special = 'REQUEST'
        # things to pop out
        queryitems = ({'key':'start', 'mkey':'start', 'type':'int'},
                      {'key':'sortorder', 'mkey':'sortorder', 'type':'string'},
                      {'key':'reverse', 'mkey':'reverse', 'type':'int'},
                      {'key':'show', 'mkey':'show', 'type':'string'},
                      )

        splitter = '-'
        if stack:
            stack_copy = copy.deepcopy(stack)
            found_item = 1
            #while found_item and stack:
            for each in range(len(stack_copy)):
                found_item = 0
                stack_item = stack_copy[each]
                for each in queryitems:
                    key, value = each['key'], each.get('mkey')

                    if value is None and stack_item==key:
                        # this is a valueless item
                        found_item = 1
                        request.set(key, 1)
                    elif stack_item.startswith("%s%s"%(key,splitter)) \
                         and value==_special:
                        found_item = 1
                        first_key = stack_item.replace("%s%s"%(key,splitter),'')
                        
                        try:
                            key, value = first_key.split(splitter,1)
                            value = int(value)
                            request.set(key, value)
                        except ValueError:
                            try:
                                key, value = first_key.split(splitter,1)
                                request.set(key, value)
                            except:
                                pass
                            
                    elif stack_item.startswith("%s%s"%(key,splitter)):
                        found_item = 1
                        if each['type']=='int':
                            replace_what = "%s%s"%(key,splitter)
                            key = int(stack_item.replace(replace_what,''))
                        else:
                            replace_what = "%s%s"%(key,splitter)
                            key = stack_item.replace(replace_what,'')
                        
                        request.set(value, key)

                if found_item:
                    stack.remove(stack_item)
                    popped.append(stack_item)

            request.set('popped',popped)

    def aurl(self, url, params={}, ignore=[]):
        """ modify the URL to include url-request-variables """
        request = self.REQUEST
        
        splitted = url.split('/')
        
        queryitems = ({'key':'start', 'mkey':'start'},
                      {'key':'sortorder', 'mkey':'sortorder'},
                      {'key':'reverse', 'mkey':'reverse'},
                      {'key':'show', 'mkey':'show'}
                      )
        
        splitter = '-'
        # Use old things
        if type(ignore)!=type([]):
            ignore = [ignore]
        keys_applied = []
        for key, value in params.items():
            keys_applied.append(key)
            if value is not None and key not in ignore:
                splitted.append("%s%s%s"%(key, splitter, value))

        # Add new things
        for each in queryitems:
            key, mkey = each['key'], each.get('mkey')
            if mkey is not None:
                if key not in keys_applied and key not in ignore and\
                request.has_key(mkey) and request[mkey] is not None:
                    splitted.append("%s%s%s"%(key, splitter, request[mkey]))
        
        return '/'.join(splitted)


    
    def getFileIconpath(self, filename):
        """ Try to find a suitable file icon """
        default = '/misc_/OFSP/File_icon.gif'
        extension = filename.lower()[filename.rfind('.')+1:]
        if ICON_ASSOCIATIONS.has_key(extension):
            return '/%s/%s'%(ICON_LOCATION,ICON_ASSOCIATIONS[extension])
        else:
            return default

    def getFileattachmentInput(self, index, initsize=40):
        """ return either a file input field or a keep option """
        request = self.REQUEST

        
        input_field = """<input size="%s" 
        name="fileattachment:list"  type="file">"""
        icon_html = """<img hspace="2" src="%s" alt="File"
        title="File size %s" border="0" />"""
        if request.has_key(TEMPFOLDER_REQUEST_KEY):
            upload_folder = request[TEMPFOLDER_REQUEST_KEY]
            # Maybe the actual folder doesn't exist any more
            if not hasattr(self.temp_folder, upload_folder):
                return input_field%initsize
            files = self.temp_folder[upload_folder].objectValues('File')
            try:
                file = files[index]
                file_src = self.getFileIconpath(file.getId())
                file_size = self.ShowFilesize(file.getSize())
                icon = icon_html%(file_src, file_size)
                
                confirm_title = "Tick if you want to keep this file attachment"
                confirm = '<input type="checkbox" checked="checked" '
                confirm += 'name="confirm_fileattachment:list" '
                confirm += 'value="%s" title="%s">'%(file.getId(),
                                                     confirm_title)

                icon = '%s<a href="%s" title="File size %s">%s%s</a>'%\
                       (confirm, file.absolute_url(), file_size, 
                        icon, file.getId())
                return icon
                
            except:
                return input_field%initsize
        else:
            return input_field%initsize
    
    def UploadTempFiles(self):
        """ Attempt to upload fileattachments to temp_folder and stick
        some information in the REQUEST """
        request = self.REQUEST
        temp_folder_id = None
        rkey = TEMPFOLDER_REQUEST_KEY

        # first, delete all unconfirmed files
        self._removeUnConfirmedFiles()
        
        if request.get(rkey, None) not in [None,'']:
            temp_folder_id = request.get(rkey)
        
        if request.has_key('fileattachment'):
            # fileattachment is a list, deal with each item
            for file in request.get('fileattachment',[]):
                if self._isFile(file):
                    if temp_folder_id is None:
                        temp_folder_id = self._generateTempFolder()

                    temp_folder = self.temp_folder[temp_folder_id]
                    filename = getattr(file, 'filename')
                    id=filename[max(string.rfind(filename, '/'),
                                    string.rfind(filename, '\\'),
                                    string.rfind(filename, ':'),
                                    )+1:]
                    if id.startswith('_'):
                        id=id[1:]
                    temp_folder.manage_addFile(id, file=file)
            # This tests whether any files were uploaded
            if temp_folder_id is not None:
                request.set(rkey, temp_folder_id)

        return "anything"
    
    security.declarePublic('_removeUnConfirmedFiles')
    def _removeUnConfirmedFiles(self):
        """ if we have a tempfolder with files that don't have a matching 
        confirm, then delete them """
        request = self.REQUEST
        rkey = TEMPFOLDER_REQUEST_KEY
        if request.get(rkey, None) not in [None,'']:
            temp_folder = self.temp_folder[request.get(rkey)]
            confirms = self._getConfirmFileattachments()
            un_upload_ids = []
            for fileid in temp_folder.objectIds('File'):
                if not fileid in confirms:
                    un_upload_ids.append(fileid)
                
            self._deleteTempFiles(temp_folder, un_upload_ids)
            
            # Anything left now?
            if len(temp_folder.objectIds('File'))==0:
                request.set(rkey, None)
                self.temp_folder.manage_delObjects([temp_folder.getId()])
            
            
    def _deleteTempFiles(self, source, ids):
        """ simply delete some files """
        source.manage_delObjects(ids)
        

    def _isFile(self, file):
        """ Check if Publisher file is a real file """
        try:
            if hasattr(file, 'filename'):
                return getattr(file, 'filename').strip() != ''
            else:
                return false
        except:
            return false
        
    security.declarePublic('_generateTempFolder')
    def _generateTempFolder(self):
        """ Create a folder in temp_folder with randomish id and return
        its id """
        root = self.temp_folder
        timestamp = str(int(self.ZopeTime()))
        randstr = self.getRandomString(length=3, numbersonly=1)
        rand_id_start = "uploadtmp-it-%s"%timestamp
        rand_id = "%s-%s"%(rand_id_start, randstr)
        while hasattr(root, rand_id):
            new_rand_str = self.getRandomString(length=3, numbersonly=1)
            rand_id = "%s-%s"%(rand_id_start, new_rand_str)

        try:
            root.manage_addFolder(rand_id)
            tempfolder = getattr(root, rand_id)
        except "Unauthorized":
            LOG(self.__class__, PROBLEM, "Could not create temporary"\
                "folder")
            
        return rand_id

    def getFileattachmentContainer(self, only_temporary=0):
        """ if TEMPFOLDER_REQUEST_KEY is set in REQUEST return folder
        object, otherwise return self. """
        request = self.REQUEST
        rkey = TEMPFOLDER_REQUEST_KEY
        if request.has_key(rkey) and request.get(rkey) is not None:
            return getattr(self.temp_folder, request[rkey])
        elif only_temporary:
            return None
        else:
            return self
        
    def nullifyTempfolderREQUEST(self):
        """ if request has tempfolder, make it None """
        request = self.REQUEST
        rkey = TEMPFOLDER_REQUEST_KEY
        if request.has_key(rkey):
            request.set(rkey, None)
            
    def SubmitIssue(self):
        """ This is the method to create an Issue Tracker Issue. It
        relies only on the REQUEST object.
        1) Check data
        2) Try to create issue
            2a) If success, RESPONSE.redirect to issue plus Thank you message
            2b) If failure, print failed data and urge to submit again
        """

        request = self.REQUEST

        failures = {}
        has_cookie = self.has_cookie
        get_cookie = self.get_cookie
        set_cookie = self.set_cookie

        # check for existance, by adding all missing properties
        # as empty instead of non-existant.
        request.set('title',request['subject'])
#        request = self._fillRequest(request)
        
        #
        # Tune the data a bit
        #

        # strip whitespace
        for property in ['title','fromname','email',
                         'url2issue','display_format']:
            request[property] = request.get(property,'').strip()
        # Special treatment needed in case STX is used upon display
        request['description'] = request.get('description','').strip()+' '

        email_cookiekey = self.getCookiekey('email')
        name_cookiekey = self.getCookiekey('name')
        display_format_cookiekey = self.getCookiekey('display_format')
        # use cookie if not else specified
        if request.get('email','')=='' and has_cookie(email_cookiekey):
            request['email'] = get_cookie(email_cookiekey)
        if request.get('fromname','')=='' and has_cookie(name_cookiekey):
            request['fromname'] = get_cookie(name_cookiekey)
        if request.get('display_format','')=='' and \
           has_cookie(display_format_cookiekey):
            request['display_format'] = get_cookie(display_format_cookiekey)

        # append the default sections if not specified
        if len(request.get('sections',[])) == 0:
            request['sections'] = self.defaultsections

        #
        # Check data
        #

        if request.get('subject','').strip() == '':
            failures[FAILURE_SUBJECT] = FAILURE_SUBJECT_DESC

        valid_emailaddress = 1
        # to prevent problems with sending mail
        if not self.ValidEmailAddress(request.get('email','')):
            valid_emailaddress = 0

        if len(failures) >0:
            return self.ExplainSubmitErrors(failures)

        #
        # Let's submit the issue!
        #

        prefix = self.issueprefix
        genid = self.generateID(self.randomid_length, prefix)

        # if these are valid, save them
        if request.get('fromname','') != '':
            set_cookie(self.getCookiekey('name'), request.get('fromname'))
        if request.get('display_format') in self.display_formats:
            set_cookie(self.getCookiekey('display_format'),
                       request.display_format)
        if valid_emailaddress:
            set_cookie(self.getCookiekey('email'), request.get('email'))


        # filter out empty item from sections
        sections_newlist=[]
        if not self.same_type(request.get('sections',''), []):
            request['sections'] = [request.get('sections')]
        for section in request.get('sections',[]):
            if section != '':
                sections_newlist.append(section)

        # add all the properties
        title           = request.get('title')
        fromname        = request.get('fromname')
        email           = request.get('email')
        url2issue       = request.get('url2issue')
        type            = request.get('type')
        urgency         = request.get('urgency')
        description     = request.get('description')
        display_format  = request.get('display_format')
        confidential    = int(request.get('confidential',0))
        hide_me         = int(request.get('hide_me',0))
        issuedate       = getNowdate()
        status          = request.get('status',self.getStatuses()[0])
        sections        = sections_newlist


        # Do the actual object adding
        cIO = self.createIssueObject
        issue = cIO(genid, request.title, status, type, urgency,
                    sections, fromname, email, url2issue,
                    confidential, hide_me, description,
                    display_format, issuedate)
        

        # catalog it
        issue.index_object()
        
        # remember it
        self.RememberAddedIssue(genid)
        
        # Also upload the fileattachments
        self._moveTempfiles(issue)

        # upload new file attachments
        self._uploadFileattachments(issue)

        # tell the people who wants to know
        self.sendAlwaysNotify(genid)

        # Where to next?
        redirect_url = '%s?NewIssue=Submitted'%(issue.absolute_url())
        redirect_url = '%s&NotYetRecent=%s'%(redirect_url, genid)
        request.RESPONSE.redirect(redirect_url)
        
    def sendAlwaysNotify(self, issueId):
        """ send out emails to those who always notify """
        always_emailstring = ', '.join(self.getAlwaysNotify())
        tosend=self.SendIssue2Friends_script(issueId, always_emailstring)
        msg, to, fr, subject = tosend
        if to is not None:
            for to_each in self.preParseEmailString(to, aslist=1):
                body = '\r\b'.join(["From: %s"%fr,"To: %s"%to_each,
                                    "Subject: %s"%subject, "",msg])
                self.sendEmail(body, to_each, fr, subject)

    def createIssueObject(self, id, title, status, type, urgency, sections,
                          fromname, email, url2issue, confidential, hide_me,
                          description, display_format, issuedate, index=0):
        """ wrap the the self._createIssueObject() method """
        if id is None or id=='':
            # create id
            id = self.generateID(self.randomid_length, self.issueprefix)
            
        if title.strip() == '':
            raise "NoSubject", "Issue has no subject line"
        
        if status.lower() not in [x.lower() for x in self.getStatuses()]:
            raise "NoStatus", "Unrecognized issue status"
        
        if type not in self.types:
            raise "NoType", "Unrecognized issue type"
        
        if urgency not in self.urgencies:
            raise "NoUrgency", "Unrecognized issue urgency"
        
        if not self.same_type(sections, []):
            raise "NotList", "Sections is not a list"
        
        if confidential not in [1,0]:
            raise "NotBoolean", "Confidential value is not boolean (1 or 0)"
        
        if hide_me not in [1,0]:
            raise "NotBoolean", "Hide_me value is not boolean (1 or 0)"
        
        if issuedate is None or issuedate =='':
            issuedate = getNowdate()
            
        # Fine, submit it
        create_method = self._createIssueObject
        return create_method(id, title, status, type, urgency,
                             sections, fromname, email, url2issue,
                             confidential, hide_me, description,
                             display_format, issuedate, index)
        
    def _uploadFileattachments(self, destination):
        """ upload all new file attachments """
        request = self.REQUEST
        
        for file in request.get('fileattachment', []):
            if self._isFile(file):
                filename = getattr(file, 'filename')
                id=filename[max(string.rfind(filename, '/'),
                                string.rfind(filename, '\\'),
                                string.rfind(filename, ':'),
                                )+1:]
                if id.startswith('_'):
                    id=id[1:]
                try:
                    destination.manage_addFile(id, file)
                except:
                    LOG(self.__class__, PROBLEM, "Could not upload file "\
                        "id=%s"%id)
                    
    security.declarePublic('_moveTempfiles')
    def _moveTempfiles(self, destination):
        """ move from temp folder to destination """
        request = self.REQUEST
        rkey = TEMPFOLDER_REQUEST_KEY
        if request.has_key(rkey):
            files_copied = []
            upload_folder_id = request.get(rkey)
            upload_folder = self.temp_folder[upload_folder_id]
            confirms = self._getConfirmFileattachments()
            for file in upload_folder.objectValues('File'):
                if file.getId() in confirms:
                    destination.manage_addFile(file.getId(),file.data)
                    files_copied.append(file.getId())
            
            upload_folder.manage_delObjects(files_copied)
            self.temp_folder.manage_delObjects([upload_folder_id])        

    def _getConfirmFileattachments(self):
        """ return the 'confirm_fileattachments' request list """
        confirms = self.REQUEST.get('confirm_fileattachment', [])
        if type(confirms) != type([]):
            confirms = [confirms]
        return confirms
    
    def _createIssueObject(self, id, title, status, type, urgency, sections,
                     fromname, email, url2issue, confidential, hide_me,
                     description, display_format, issuedate, index=0):
        """ crudely create issue object. No checking """
        issue = IssueTrackerIssue(id, title, status, type, urgency,
                                  sections, fromname, email, url2issue,
                                  confidential, hide_me, description,
                                  display_format, issuedate)

        # not here
        self._setObject(id, issue)
        issue = getattr(self, id)

        if index:
            # catalog it
            issue.index_object()

        return issue
    
    def same_type(self, one, two):
        """ use this because 'type' as variable can be used elsewhere """
        return type(one)==type(two)

    def generateID(self, length, prefix='',
                    meta_type=ISSUE_METATYPE,increment=None):
        """ generate IDs for different objects """
        if increment is None:
            idnr = len(self.objectIds(meta_type))+1
            increment = idnr + 1
        else:
            idnr = increment
            increment = increment +1
        id='%s%s'%(prefix, string.zfill(str(idnr), length))
        base= getattr(self,'aq_base',self)
        if hasattr(base, id):
            # ah! Id already exists, try again
            return self.generateID(length, prefix, meta_type, 
                                   increment=increment)
        else:
            return id


    def About(self):
        """ Show some info about the product """
        osj = os.path.join
        About_content = open(osj(package_home(globals()), 'About.html')).read()
        f = open(osj(package_home(globals()), 'CHANGES.txt'), "r")
        changelog = f.read()
        f.close()
        changelog = self.ShowDescription(changelog.strip()+' ', 'structuredtext')
        About_content = About_content.replace('*CHANGELOG*',changelog)
        About_content = About_content.replace('*VERSION*',__version__)
        return About_content
    
    def RSS091(self, batchsize=None):
        """ return RSS XML """
        request = self.REQUEST
        root = self.getRoot()
        xml="""<?xml version="1.0"?><rss version="0.91">
        <channel>
        <title>%s</title>
        <link>%s</link>
        <description>%s</description>
        <language>en-uk</language>
        <copyright></copyright>
        <webMaster>%s</webMaster>\n"""%\
           (root.title, root.absolute_url(), root.title,
            self.sitemaster_email)
        logo = getattr(self, 'issuetracker_logo.gif')
        xml=xml+"""<image>
        <title>%s</title>
        <url>%s</url>
        <link>%s</link>
        <width>%s</width>
        <height>%s</height>
        <description>%s</description>
        </image>\n"""%(logo.title, logo.absolute_url().strip(),
                       root.absolute_url(),
                       logo.width, logo.height,
                       root.title)
        # manually set sortorder
        request.set('sortorder','date')
        request.set('reverse',1)
        if batchsize is None:
            batchsize = self.default_batch_size
        for issue in self.ListIssuesFiltered()[:batchsize]:
            title = "%s (%s)"%(issue.title, issue.status.capitalize()) 
            title = self._prepare_feed(title)
            description = self._prepare_feed(issue.description)

            xml=xml+"""\n\t<item>
            <title>%s</title>
            <description>%s</description>
            <link>%s</link>
            """%(title, description, issue.absolute_url())
            if issue.fromname != '':
                author = "%s (%s)"%(issue.fromname, issue.email)
                xml="%s\n<author>%s</author>\n"%(xml, author)
            xml=xml+"\n\t</item>"
            
        xml=xml+"""</channel>\n</rss>"""
        
        response = request.RESPONSE
        response.setHeader('Content-Type', 'text/xml')
        return xml
    
    def _prepare_feed(self, s):
        """ prepare the text for XML usage """
        _replace = IssueTrackerUtils._replace_special_chars
        s = html_quote(s)
        s = s.replace('','&#163;')
        s = _replace(s, html_encoding=1)
        s = s.replace('&','&amp;')
        return s

    # Cookies!

    def get_cookie(self, name, default=None):
        """ return RESPONSE cookie """
        return self.REQUEST.cookies.get(name,default)
    
    def set_cookie(self, name, value):
        """ set RESPONSE cookie """
        later = str(IssueTrackerUtils.getLaterDate(365))
        self.REQUEST.RESPONSE.setCookie(name,value, expires=later, path='/')
        
    def has_cookie(self, name):
        """ return cookie presence """
        return self.REQUEST.cookies.has_key(name)

    # Sessions!
    
    def get_session(self, name, default=None, unique=0):
        """ Override the session.get method a little bit """
        if not unique:
            name = self.defineInstanceCookieKey(name)
        return self.REQUEST.SESSION.get(name, default)
    
    def set_session(self, name, value, unique=0):
        """ Overrode the session.set method a little bit """
        if not unique:
            name = self.defineInstanceCookieKey(name)
        self.REQUEST.SESSION.set(name, value)
        
    def has_session(self, name, unique=0):
        """ Override the session.has_key method a little big """
        if not unique:
            name = self.defineInstanceCookieKey(name)
        return self.REQUEST.SESSION.has_key(name)

    
    # Misc

    def showURL2Issue(self, url=None, href=0):
        """ display the url2issue for ShowIssueData """
        if url is None:
            url = self.url2issue
        if href:
            if not (url.startswith('http://') or url.startswith('https://')):
                url = 'http://'+url
            return url
        else:
            url = url.replace('http://','')
            if len(url) > 40:
                url = url[0:20]+'...'+url[-20:]
            return url
    
    def displayBriefTitle(self, title):
        """ return the title or truncate it a bit """
        if len(title) > BRIEF_TITLE_MAX_LENGTH:
            title = title[:BRIEF_TITLE_MAX_LENGTH]+'...'
        return title
    
    def url_quote_plus(self, s):
        """ wrap the url_quote_plus script from standards """
        return url_quote_plus(s)
    
    def UpdateCatalog(self, REQUEST=None):
        """ Re-find items in the Catalog """
        request = self.REQUEST

        if request.has_key('URL2'):
            URL2 = request.URL2
        else:
            URL2 = request.URL1

        #method = self.getCatalog().manage_catalogFoundItems
        #method(request, request.RESPONSE, URL2, request.URL1,
               #obj_metatypes=[ISSUE_METATYPE, ISSUETHREAD_METATYPE])

        catalog = self.getCatalog()
        catalog.manage_catalogClear()
        
        for issue in self.getRoot().objectValues(ISSUE_METATYPE):
            issue.index_object()
            for thread in issue.objectValues(ISSUETHREAD_METATYPE):
                thread.unindex_object()
                thread.index_object(idxs=['id','comment'])

        msg = "%s updated."%catalog.getId()
        if REQUEST is None:
            return msg
        else:
            method = IssueTrackerUtils.AddParam2URL
            desturl = self.getRootURL()+"/manage_ManagementForm"
            url = method(desturl,{'manage_tabs_message':msg})
            self.REQUEST.RESPONSE.redirect(url)
                
        
    def issueURLbyID(self, issueID):
        """ Return absolute_url of an issue from its id """
        return getattr(self.getRoot(),issueID).absolute_url()

    def getAllNotifications(self):
        """ Go through all issues and find all notification objects """
        all = []
        for issue in self.getRoot().objectValues(ISSUE_METATYPE):
            all = all+issue.objectValues('Issue Notification')
        return all
    
    def LineIndent(self, text, indent):
        """ wrap script """
        return IssueTrackerUtils.LineIndent(text, indent)
    
    def preParseEmailString(self, email_string,
                            aslist=0, allnotifyables=1):
        """ wrapper around utils """
        parsemethod = IssueTrackerUtils.preParseEmailString
        all_notifyables = self.getNotifyables()
        if not allnotifyables:
            all_notifyables = []
        if all_notifyables:
            names2emails = {}
            for item in all_notifyables:
                names2emails[item.getName()] = item.getEmail()
            result = parsemethod(email_string, names2emails=names2emails,
                                 aslist=aslist)
            
        else:
            result = parsemethod(email_string, aslist=aslist)
        return result
       

    def getManagerRoles(self):
        """ Return the roles that makes an IssueTracker Manager """
        return getattr(self, 'manager_roles', DEFAULT_MANAGER_ROLES)

    def hasManagerRole(self):
        """
            This method determines if the current user
            is allowed to do stuff that only the Zope manager is
            supposed to be able to do.
            Feel free to edit appropriatly to what suits you.
        """
        user_roles = self.REQUEST.AUTHENTICATED_USER.getRoles() 
        for role in self.getManagerRoles():
            if role in user_roles:
                return true
        # still here!
        return false

    def showFilterOptions(self):
        """ Determine if we want to display the filter options """
        request = self.REQUEST
        
        showkey = SHOW_FILTEROPTIONS_KEY
        key = FILTEROPTIONS_KEY
        rkey = 'ShowFilterOptions'
        if request.has_key(rkey) and int(request[rkey]):
            # Someone has chosen to show filter options
            self.set_session(showkey, 1)
            return true
        elif request.has_key(rkey) and not int(request[rkey]):
            self.set_session(showkey, 0)
            return false
        elif self.has_session(key) and self.get_session(key) or\
        self.has_session(showkey) and self.get_session(showkey):
            return true
        else:
            return false
        
    def AllowIssueAttributeChange(self):
        """ Determine if the allow_issueattrchange is true """
        return getattr(self, 'allow_issueattrchange',
                       DEFAULT_ALLOW_ISSUEATTRCHANGE)

    def AllowIssueSubscription(self):
        """ Determine if the allow_subscription is true """
        return getattr(self, 'allow_subscription', DEFAULT_ALLOW_SUBSCRIPTION)
        
    def PrivateStatistics(self):
        """ Determine if private_statistics is false """
        default = DEFAULT_PRIVATE_STATISTICS
        return getattr(self, 'private_statistics', default)
    
    def ShowConfidentialOption(self):
        """ return show_confidential_option """
        default = DEFAULT_SHOW_CONFIDENTIAL_OPTION
        return getattr(self, 'show_confidential_option', default)
    
    def ShowHideMeOption(self):
        """ return show_hideme_option """
        default = DEFAULT_SHOW_HIDEME_OPTION
        return getattr(self, 'show_hideme_option', default)
    
    def ShowDownloadButton(self):
        """ return show_download_button """
        default = DEFAULT_SHOW_DOWNLOAD_BUTTON
        return getattr(self, 'show_download_button', default)
    
    def EncodeEmailDisplay(self):
        """ return encode_emaildisplay """
        default = DEFAULT_ENCODE_EMAILDISPLAY
        return getattr(self, 'encode_emaildisplay', default)
    
    def getRandomString(self, length=5, loweronly=0, numbersonly=0):
        """ return a completely random piece of string """
        script = IssueTrackerUtils.getRandomString
        return script(length, loweronly, numbersonly)

    
    def thisInURL(self, page, homepage=0):
        """ To find if a certain objectid is in the URL """
        URL = self.ActionURL(self.REQUEST.URL)
        rootURL = self.getRootURL()
        if homepage and URL==rootURL:
            return true
        else:
            expected = rootURL +'/'+ page
            if URL.lower() == expected.lower():
                return true
            else:
                return false

    def ActionURL(self, url=None):
        """
        If URL is http://host/index_html
        I prefer to display it http://host
        Just a little Look&Feel thing
        """
        if url is None:
            url = self.REQUEST.URL

        URLsplitted = string.split(url,'/')
        if URLsplitted[-1] == 'index_html':
            return string.join(URLsplitted[:-1],'/')

        return url

    def ExplainSubmitErrors(self, dict):
        """ Return HTMLed statements about errors (naughty)"""
        li = []
        for key,value in dict.items():
            li.append('<strong>%s</strong> - %s'%(key, value))
        return li

    def ShowFilter(self, filtername, sequence=[]):
        """ Check whether to show filter or not """
        request = self.REQUEST
        key = FILTEROPTIONS_KEY
        if request.has_key(filtername):
            return request[filtername]
        elif self.has_session(key):
            filteroptions = self.get_session(key)
            if filteroptions.has_key(filtername):
                return filteroptions[filtername]
        return []

    def hasStoredFilter(self):
        """ Check if filter is applied """
        key = FILTEROPTIONS_KEY
        return self.has_session(key) and self.get_session(key)


    def ShowOnlyMe(self):
        """ enable listing with set name and email """
        request = self.REQUEST
        redirect = request.RESPONSE.redirect

        filters={}
        for each in ['fromname','email']:
            if request.has_key(each) and request[each] != '':
                filters[each]=request[each]

        if filters:
            querystring = []
            for k,v in filters.items():
                querystring.append('%s=%s'%(k,v))
            # make list into string
            querystring = '?'+string.join(querystring,'&')
        else:
            querystring=''

        listpage = '/%s'%self.whichList()
        redirect(self.getRootURL()+listpage+querystring,lock=1)

    def EmailHyperlink(self, email, title=None):
        """ if encode_emaildisplay then use JavaScript to encode it """
        if title is None:
            title = email
           
        if self.EncodeEmailDisplay():
            js_string = """document.write('<a href="mailto:%s">"""%email
            js_string += """%s</a>')"""%title
            hexed = self._hex_string(js_string)
            js_script = """<script language="JavaScript">eval(unescape('"""
            js_script += hexed + """'))</script>"""
            return js_script
        else:
            return '<a href="mailto:%s">%s</a>'%(email, title)
        
            
    def _hex_string(self, oldstring):
        """ hexify a string """
        # Taken from http://www.happysnax.com.au/testemail.php
        # See Credits
        
        def _tohex(n):
            hs='0123456789ABCDEF'
            return hs[int(floor(n/16))]+hs[n%16]
        
        newstring=''
        length=len(oldstring)
        for i in range(length):
            newstring=newstring+'%'+_tohex(ord(oldstring[i]))
        return newstring
    


    def ShowNameEmail(self, fromname, email, hideme=None, highlight=1):
        """ Show name and email depending on certain criterias """
        out = ''
        if highlight:
            fromname = self.HighlightQ(html_quote(fromname))
            email    = html_quote(email)
            show_email = self.HighlightQ(email)
            
        else:
            fromname = html_quote(fromname)
            email    = html_quote(email)
            show_email = email

        if fromname =='' and email=='':
            name_email = NONAME_NOEMAIL
        elif fromname =='': 
            hyperlink = self.EmailHyperlink(email, show_email)
            name_email ='&lt;%s&gt;'%hyperlink
        elif email =='':
            name_email = fromname
        else:
            hyperlink = self.EmailHyperlink(email, show_email)
            name_email = '%s &lt;%s&gt;'
            name_email = name_email%(fromname, hyperlink)
         

        if hideme is not None and hideme:
            out += NAME_EMAIL_HIDDEN 
            if self.hasManagerRole():
                out += "<br>" + name_email
        else:
            out += name_email

        return out

    def getIssueURL(self, issue, no_q=0):
        """ Return an issues URL but maybe add a thing to it """
        url = issue.absolute_url()
        request = self.REQUEST
        if no_q or not (request.has_key('q') and request['q']!=''):
            return url
        else:
            return "%s?q=%s"%(url, url_quote_plus(request['q']))

    def setFollowupCursor(self):
        """ Check if it's time to add a comment, and if that's the case
        prepare something in the REQUEST so that StandardLook can pick
        up on that. """
        request = self.REQUEST
        if self.has_key_special('IssueAction', 1) and \
           not request.has_key('IssueAction_Delete'):
            textarea_name = "document.form_followup['comment_description:text']"
            self.REQUEST.set('FirstFocus', textarea_name)
            
    def setAddIssueCursor(self):
        """ If it's addIssue or quick, set the focus """
        textarea_name = "document.add_issue.subject"
        if not self.REQUEST.has_key('previewissue'):
            self.REQUEST.set('FirstFocus', textarea_name)

    def getFirstFocusJavascript(self):
        """ create a script tag if 'FirstFocus' """
        request = self.REQUEST
        s="function setFirstFocus() {"
        if request.has_key('FirstFocus'):
            s=s+ request.get('FirstFocus')+'.focus();'
        s=s+'}'
        return s
        
    def has_key_special(self, name, shorten=0):
        """
        Normally you would do REQUEST.has_key('IssueAction')
        but if an imagebutton is used you'll find that you have
        REQUEST['IssueAction.y'] and REQUEST['IssueAction.x']
        But the result should be the same.
        """
        request = self.REQUEST
        if request.has_key(name):
            return true
        elif request.has_key('%s.y'%name) and request.has_key('%s.x'%name):
            return true
        elif shorten:
            for key in request.keys():
                if key[:len(name)]==name:
                    return true
            return false
        else:
            return false

    def get_special_key(self, name):
        """
        Normally you would do REQUEST.has_key('IssueAction')
        but if an imagebutton is used you'll find that you have
        REQUEST['IssueAction.y'] and REQUEST['IssueAction.x']
        But the result should be the same.
        """
        try:
            return self.REQUEST[name]
        except KeyError:
            try:
                return self.REQUEST['%s.x'%name]
            except:
                raise KeyError, name

    def getSavedUser(self, name_email='email'):
        """
        Return the name or email from request,
        if not found, return from cookie,
        else return ""
        """
        request = self.REQUEST

        if name_email =='email':
            s='email'
            cookie=self.getCookiekey('email')
        else:
            s='fromname'
            cookie=self.getCookiekey('name')

        # now we know what we're looking for

        if request.has_key(s):
            return request[s]
        elif self.has_cookie(cookie):
            return self.get_cookie(cookie)

        return ""

    def getPreviewTitle(self, oldstatus, action):
        """ Get what the title of thread will be (via the web only) """
        request = self.REQUEST
        action = ss(action)
        addfollowup = 0
        past_tense=None
        
        statuses_and_verbs = self.getStatusesMerged(asdict=1)
        lowercase_values = {}
        statuses_and_verbs_reversed = {}
        for key, value in statuses_and_verbs.items():
            lowercase_values[value.lower()] = value
            statuses_and_verbs_reversed[value] = key
            
        if lowercase_values.has_key(action):
            past_tense = statuses_and_verbs_reversed[lowercase_values[action]]
            addfollowup = 1
            
        #if action == 'open':
            #past_tense='open'
            #addfollowup = 1
        #if action == 'take':
            #past_tense='taken'
            #addfollowup = 1
        #if action == 'put on hold':
            #past_tense='on hold'
            #addfollowup = 1
        #if action == 'reject':
            #past_tense='rejected'
            #addfollowup = 1
        #if action == 'complete':
            #past_tense='completed'
            #addfollowup = 1
        elif action == 'delete':
            DeleteIssue(self, self.id)
            addfollowup = 0
            redirect_url = request.URL2
        elif action == 'addfollowup':
            addfollowup = 1
            dummychange = 1
            req_manager = 0
            
        if action == 'addfollowup':
            gentitle = 'Added Issue followup'
        else:
            gentitle = 'Changed status from %s to %s'%\
                       (oldstatus.capitalize(), past_tense.capitalize())

        return gentitle


    def ManagerOptionsExtend(self):
        """ Determine which forms to show to Managers """
        request = self.REQUEST

        has_key_special = self.has_key_special
        get_special_key = self.get_special_key

        do_preview = 0
        if has_key_special('IssueAction') and has_key_special('previewissue'):
            do_preview = 1
            
        form = ''
        if has_key_special('IssueAction') and not do_preview:
            action = get_special_key('IssueAction')
            return getattr(self,request['issueID']).ModifyIssue(action=action)
        elif has_key_special('IssueAction_AddFollowup'):
            request.set('no_quick_formfollowup',1)
            return self.form_followup()
        elif has_key_special('IssueAction_Delete'):
            return self.form_delete()
        else:
            issuestatus = self.status.lower()
            for item in self.getStatusesMerged(aslist=1):
                status, verb = item
                action = 'IssueAction_%s'%verb.replace(' ','').capitalize()
                if has_key_special(action):
                    otherTitle = 'Change status from '
                    otherTitle += '<em>%s</em> '%issuestatus.capitalize()
                    otherTitle += 'to <em>%s</em>'%status.capitalize()
                    request.set('otherActionTitle', otherTitle)
                    request.set('otherAction', verb.capitalize())
                    request.set('otherComment', 'Optional comment')
                    request.set('whatAction', action)
                    request.set('no_quick_formfollowup',1)
                    return self.form_followup()

            else:
                return self.manager_options()
        #elif has_key_special('IssueAction_Open'):
            #form = self.form_open()
        #elif has_key_special('IssueAction_Take'):
            #form = self.form_take()
        #elif has_key_special('IssueAction_Onhold'):
            #form = self.form_onhold()
        #elif has_key_special('IssueAction_Rejected'):
            #form = self.form_reject()
        #elif has_key_special('IssueAction_Complete'):
            #form = self.form_complete()
        #elif has_key_special('IssueAction_Delete'):
            #form = self.form_delete()
        #else:
            #form = self.manager_options()

        #return form

    def AnonymousOptionsExtend(self):
        """ Determine which forms to show to Anonymous """
        request = self.REQUEST

        has_key_special = self.has_key_special
        get_special_key = self.get_special_key

        do_preview = 0
        if has_key_special('IssueAction') and has_key_special('previewissue'):
            do_preview = 1

        form = ''
        if has_key_special('IssueAction') and not do_preview:
            action = get_special_key('IssueAction')
            return getattr(self,request['issueID']).ModifyIssue(action=action)
        elif has_key_special('IssueAction_AddFollowup'):
            form = self.form_followup()
            request.set('no_quick_formfollowup',1)
        else:
            form = self.anonymous_options()

        return form


    def CountStatuses(self, since=None):
        """ Return how many Issues there are under each status since
            a certain time
        """
        #
        # NEEDS WORK!!
        #
        
        #return {}
        request = self.REQUEST

        # check that since isn't a string
        if type(since)==StringType:
            since = DateTime(since)
        elif request.has_key('count_status_since'):
            since = DateTime()-int(request['count_status_since'])


        res={}
        tres=[]
        for issue in self.objectValues(ISSUE_METATYPE):
            if since is None or issue.bobobase_modification_time() >= since:
                status = issue.status.lower()
                if res.has_key(status):
                    res[status]=res[status]+1
                else:
                    res[status] = 1

        # Lastly we want to organize res by self.statuses order
        for status in self.getStatuses():
            status = status.lower()
            if res.has_key(status):
                #sc = StatusCount(status, res[status])
                tres.append([status, res[status]])
            else:
                #sc = StatusCount(status)
                tres.append([status, 0])
        return tres

    def BatchedQueryString(self, batchdict={}):
        """
        return QUERY_STRING but make sure stuff in
        the batchdict isn't duplicated.
        """
        request = self.REQUEST
        actionurl = self.ActionURL()
        if type(batchdict)==type('s') and batchdict=='all':
            #request.set('start', None)
            url = self.aurl(actionurl, {'show':'all'}, ignore='start')
        elif type(batchdict)==type('s') and batchdict.lower()=='none':
            url = self.aurl(actionurl, ignore=['start','show'])
        else:
            batchdict = self._Zero2None(batchdict)
            url = self.aurl(actionurl, batchdict)
        url = self._addQuerystring(url)
        return url
    

    def _Zero2None(self, dict):
        """ Replace all occurances of 0 (as tested int) to None """
        n_dict={}
        for key, value in dict.items():
            try:
                if int(value)==0:
                    n_dict[key]=None
                else:
                    n_dict[key]=value
            except:
                n_dict[key]=value
        return n_dict
    
    def DeleteIssue(self):
        """ Delete an Issue from the IssueTracker instance """
        request = self.REQUEST

        if request.has_key('issueID') and self.hasManagerRole():
            self.manage_delObjects(request['issueID'])

            # delete all notifications about this Issue
            del_notify_ids = []
            for notifyobject in self.objectValues('Issue Notification'):
                if notifyobject.issueID == request['issueID']:
                    del_notify_ids.append(notifyobject.id)
          
            self.manage_delObjects(del_notify_ids)

            listpage = '/%s'%self.whichList()
            request.RESPONSE.redirect(request.URL1+listpage)
        else:
            msg = "The issueID could not be found in the REQUEST"
            raise "IssueTrackerError", msg

    def standard_html_header(context):
        """ to make it possible to use DTML objects here """
        breakword = '<!--METALbody-->'
        page = context.StandardLook()
        return page[:page.find(breakword)]

    def standard_html_footer(context):
        """ to make it possible to use DTML objects here """
        breakword = '<!--METALbody-->'
        page = context.StandardLook()
        return page[page.find(breakword)+len(breakword)+1:]

    def ManagerLink(context):
        """ For the little hyperlink where you can login with """
        link = context.login.absolute_url()
        came_from = context.ActionURL()+'/'
        rurl=str(int(getNowdate()))[-4:-1]
        return "%s?came_from=%s&r.url=%s"%(link, came_from, rurl)

    def getSavedTextFormat(self, no_default=0):
        """
        This method returns what display_format value the user has. 
        If none found, then the default one is returned.
        """

        if no_default:
            default = ""
        else:
            default = self.default_display_format

        s=None
        cookiekey = self.getCookiekey('display_format')

        if self.has_cookie(cookiekey):
            s = self.get_cookie(cookiekey)

        if s not in self.display_formats:
            s = None

        if s is None:
            return default
        else:
            return s

    def getNoFileattachments(self):
        """ return no_fileattachments or default """
        return getattr(self, 'no_fileattachments', DEFAULT_NO_FILEATTACHMENTS)
    
    def getNoFollowupFileattachments(self):
        """ return no_followup_fileattachments or default """
        return getattr(self, 'no_followup_fileattachments',
                       DEFAULT_NO_FOLLOWUP_FILEATTACHMENTS)

    def getAlwaysNotify(self):
        """ return always_notify or default """
        always = getattr(self, 'always_notify', DEFAULT_ALWAYS_NOTIFY)
        return self._filter_emptylistitems(always)

    def _filter_emptylistitems(self, l):
        nl=[]
        for i in l:
            if i.strip()!='':
                nl.append(i)
        return nl
        
    def saveEmailstring(self, to):
        """ Save to string as a cookie """
        key = EMAILSTRING_COOKIEKEY
        key = self.defineInstanceCookieKey(key)
        self.set_cookie(key, to)

    def getSavedEmailstring(self):
        """ Return cookie translated or nothing """
        key = EMAILSTRING_COOKIEKEY
        key = self.defineInstanceCookieKey(key)
        
        if self.REQUEST.cookies.has_key(key):
            to = self.REQUEST.cookies[key]
            for item in self.getNotifyables():
                to = to.replace(item.getEmail(), item.getName())
            return to
        else:
            return None

    def sendEmail(self, body, to, fr, subject):
        """ Attempt to send emails un protectedly """
        if hasattr(self, 'MailHost'):
            mailhost = self.MailHost
        else:
            all_mailhosts = self.superValues('Mail Host')
            if all_mailhosts:
                mailhost = all_mailhosts[0] # first one
            else:
                raise "AttributeError", "MailHost object not found"
        try:
            mailhost.send(body, to, fr, subject)
        except:
            type, val, tb = sys.exc_info()
            LOG(self.__class__.__name__, WARNING,
                'Could not send email to %s'%to, '')
            raise type, val
            
    

    
    # List issues
    #

    def setWhichList(self, what):
        """ set a SESSION with which list """
        key = WHICHLIST_SESSION_KEY
        session = self.REQUEST.SESSION
        what = what.lower().strip()
        if what in ['completelist','listissues']:
            try:
                session.set(key, what)
            except:
                # tough!
                pass
        return None
    
    def whichList(self):
        """ inspect the SESSION object if there's information
        about either "ListIssues" or "CompleteList"
        """
        session = self.REQUEST.SESSION
        key = WHICHLIST_SESSION_KEY
        default = 'ListIssues'
        if session.get(key,'').lower().strip() == 'completelist':
            return 'CompleteList'
        else:
            return default

    def ResetFilter(self, page='ListIssues'):
        """ reset the filter then show the ListIssues or eq. again """
        key = FILTEROPTIONS_KEY
        self.set_session(key, {})
        page = page.lower().strip()
        if page == 'listissues':
            page = '/ListIssues'
        elif page == 'completelist':
            page = '/CompleteList'
        else:
            raise "NotFound"
        self.REQUEST.RESPONSE.redirect(self.getRootURL()+page)
        
    def HideFilter(self, page='ListIssues'):
        """ hide the filter then show the ListIssues or eq. again """
        key = SHOW_FILTEROPTIONS_KEY
        self.set_session(key, false)
        page = page.lower().strip()
        if page == 'listissues':
            page = '/ListIssues'
        elif page == 'completelist':
            page = '/CompleteList'
        else:
            raise "NotFound"
        self.REQUEST.RESPONSE.redirect(self.getRootURL()+page)
        
    
    def ListIssuesFiltered(self):
        """ wrapper around _ListIssuesFiltered() that prepares a search
        if REQUEST holds 'q'
        """
        request = self.REQUEST
        q=None
        catalog = self.getCatalog()

        root = self.getRoot()
        if request.has_key('q') and request['q'].strip() != '':
            q=request['q'].strip()
            transtab = string.maketrans('/ ','_ ')
            q=string.translate(q, transtab, '?&!;()<=>*#[]{}')
##            q=q.replace('%','*') # allow both wildcards # needs thought
        seq = []
        if q is not None:
            catalogs = catalog.searchResults(title='*%s*'%q) + \
                       catalog.searchResults(description=q) + \
                       catalog.searchResults(fromname=q) + \
                       catalog.searchResults(email=q)            
            
            catalogs = self._filter_catalog(catalogs)
            
            # Now, also search on comment
            catalogs_threads = catalog.searchResults(comment=q)
            catalogs_threads = self._filter_catalog(catalogs_threads, threads=1)
            
            if len(catalogs)+len(catalogs_threads)==0:
                # nothing found, maybe user typed in an id
                catalogs = self._filter_catalog(catalog({'id':q}))
                if len(catalogs)==1:
                    response = request.RESPONSE
                    url = catalogs[0].getObject().absolute_url()
                    response.redirect(url)
                
        else:
            catalogs = self._filter_catalog(catalog())
            catalogs_threads = []

        for brain in catalogs:
            object = brain.getObject()
            if object not in seq:
                seq.append(object)

        if q is not None:
            # Also, search the file attachments
            for issue in self._findby_filename(q):
                if issue not in seq:
                    seq.append(issue)
            
        first_thread_id = None
        for threadbrain in catalogs_threads:
            threadobject = threadbrain.getObject()
            object = threadobject.aq_parent
            if object not in seq:
                if first_thread_id is None:
                    first_thread_id = object.getId()
                    request.set('FirstThreadResultId', first_thread_id)
                seq.append(object)
        if q is not None:
            # Remember this searchterm
            self.RememberSearchTerm(q, len(seq))
                
        return self._ListIssuesFiltered(seq)
        
    def _filter_catalog(self, catalogs, threads=0):
        """ Make sure the issues in the catalogs is only in this issue tracker """
        root = self.getRoot()
        root_issues = root.objectValues(ISSUE_METATYPE)
        n_catalogs=[]
        for brain in catalogs:
            if brain.getObject() in root_issues:
                n_catalogs.append(brain)
            elif threads and brain.getObject().aq_parent in root_issues:
                n_catalogs.append(brain)
        return n_catalogs

    def _findby_filename(self, q):
        """ Search all file attachments """
        q = q.lower()
        issues = []
        r = self.ZopeFind(self, obj_metatypes=['File'], search_sub=1)
        valid_meta_types = [ISSUE_METATYPE, ISSUETHREAD_METATYPE]
        for file in r:
            path, fileobject = file
            parent = fileobject.aq_parent
            if parent.meta_type in valid_meta_types and path.lower().find(q)>-1:
                if parent.meta_type == ISSUETHREAD_METATYPE:
                    issues.append(parent.aq_parent)
                else:
                    issues.append(parent)
                
        return issues
    
    def _ListIssuesFiltered(self, objsequence):
        """ Basically filter a sequence of issue objects """
        rlist = []
        request = self.REQUEST
        rget = request.get
        rset = request.set
        rset('TotalNoIssues', len(objsequence))
        ckey = self.getCookiekey('sortorder')
        crkey = self.getCookiekey('sortorder_reverse')
        
        if request.has_key('sortorder') and rget('sortorder') is None:
            rset('sortorder','ignore')
            
        if rget('sortorder','').lower()=='search' and rget('q','').strip() != '':
            search_sort = true
        else:
            search_sort = false
            
        if request.has_key('reverse'):
            reverse = int(request['reverse'])
            if not rget('sortorder','')=='ignore':
                self.set_cookie(crkey, reverse)
        elif not request.has_key('sortorder') and \
             int(request.cookies.get(crkey, 0)):
            reverse = 1
            rset('reverse', 1)
        else:
            reverse = 0
            if not rget('sortorder','')=='ignore':
                self.set_cookie(crkey, reverse)
                
            
        # remember filteroptions
        key = FILTEROPTIONS_KEY
        if request.has_key('filteroptions'):
            tracker_filteroptions={}
            for each in ['ignoretypes','ignoresections','ignorestatuses',
                         'ignoreurgencies','ignorefromname','ignoreemail',
                         'showfromname','showemail']:
                try:
                    tracker_filteroptions[each]=request[each]
                except:pass
            
            self.set_session(key, tracker_filteroptions)    
            
       
        #
        # prepare a master filter
        #
        
        # Convert data from session into request
        filteroptions = self.get_session(key, {})
        for each in ['ignorestatuses','ignoresections', 'ignoreurgencies', 
                     'ignoretypes','showfromname','showemail']:
            if filteroptions.has_key(each):
                rset(each, filteroptions[each])

        filter = {}
        
        # statuses
        if request.has_key('ignorestatuses'):
            filter['status'] = request.ignorestatuses
        elif request.has_key('status'):
            filter['status'] = self._pop_from_list(request['status'],
                                                   self.getStatuses())
        else:
            filter['status'] = []
            
        # sections
        if request.has_key('ignoresections'):
            filter['sections'] = request.ignoresections
        elif request.has_key('section'):
            filter['sections'] = [request.section.strip()]
        else:
            filter['sections'] = []

        # urgencies
        if request.has_key('ignoreurgencies'):
            filter['urgency'] = request.ignoreurgencies
        elif request.has_key('urgency'):
            filter['urgency'] = [request.urgency.strip()]
        else:
            filter['urgency'] = []
            
        # types
        if request.has_key('ignoretypes'):
            filter['type'] = request.ignoretypes
        elif request.has_key('type'):
            filter['type'] = [request.type.strip()]
        else:
            filter['type'] = []
            
        ## fromname
        # Either a name to ignore or show only.
        # It does not make sense to have both showfromname and ignorefromname
        if request.has_key('showfromname') and \
           request.showfromname <> self.when_ignore_word:
                filter['showfromname'] = request.showfromname.strip()
        elif request.has_key('ignorefromname') and \
             request.ignorefromname <> self.when_ignore_word:
            filter['ignorefromname'] = request.fromname.strip()

        # email
        if request.has_key('showemail') and \
           request.showemail <> self.when_ignore_word:
                filter['showemail'] = string.strip(request.showemail)
        elif request.has_key('ignoreemail') and \
           request.ignoreemail <> self.when_ignore_word:
            filter['ignoreemail'] = request.email


        # prepare for caseindependence
        filter = fixDictofLists(filter)
            
        # all Issue objects
        issues = objsequence 

        #
        # Before we can determine what to sort on
        # we need to check the name of the properties
        
        # the default
        defaultdefault = 'bobobase_modification_time'
        sortorder = (defaultdefault,)
        if request.has_key('sortorder') and not rget('sortorder') in ['search','ignore']:
            self.set_cookie(ckey, rget('sortorder'))
        else:
            # Use cookie instead?
            if self.has_cookie(ckey) and not rget('sortorder','')=='ignore':
                if not search_sort or rget('sortorder','')=='search' and rget('q','').strip()=='':
                    rset('sortorder', request.cookies.get(ckey))
        skipsort = 0
        if request.has_key('sortorder') and request['sortorder'] != 'date':
            sortorder = (rget('sortorder'),)
            if rget('sortorder') == 'from':
                sortorder = ('fromname',)
            elif rget('sortorder') == 'creationdate':
                sortorder = ('issuedate',)
            elif rget('sortorder') in ['changedate','ignore']:
                sortorder = ('bobobase_modification_time',)

            if rget('sortorder') == 'status':
                #
                # Very special!
                # self.getStatuses() is an "human ordered" list like 
                # 'open','taken',...'completed','forgotten'
                statuses = {}
                # create a dictionary of list divided in groups by their status
                # like this: 
                # {'open':['foo','bar'], ... 'forgotten':['this','that']}
                for issue in issues:
                    if statuses.has_key(issue.status):
                        statuses[issue.status].append(issue)
                    else:
                        statuses[issue.status] = [issue]
                # recreate the containers
                issues = []
                # remembers that self.statuses is ordered
                #
                # self.statuses is the property
                # statuses is the dictionary
                #
                for status in self.getStatuses():
                    # perhaps there are no 'forgotten' issues even
                    # though there is 'forgotten' in self.statuses
                    if statuses.has_key(status):
                        # order all issues with this status 
                        # by bobobase_modification_time
                        thisstatus_containers = self._dosort(statuses[status], 
                                                             defaultdefault)
                        # reverse bobobase_modification_time sort naturally
                        thisstatus_containers.reverse()
                        issues = issues + thisstatus_containers
                        # skip the default sort later
                        skipsort = 1
                        
            if request['sortorder'] == 'urgency':
                #
                # Very special too
                # read the comments on "if request['sortorder'] == 'status':"
                # above.
                urgencies = {}
                for issue in issues:
                    if urgencies.has_key(issue.urgency):
                        urgencies[issue.urgency].append(issue)
                    else:
                        urgencies[issue.urgency] = [issue]
                issues = []
                for urgency in self.urgencies:#+():
                    if urgencies.has_key(urgency):
                        thisurgency_containers = self._dosort(urgencies[urgency], defaultdefault)
                        thisurgency_containers.reverse()
                        issues = issues + thisurgency_containers
                        skipsort = 1
                
        else:
            # if no sortorder, it uses the bobobase_modification_time
            # and this has to be reversed so that the latest is listed first
            reverse = 1

        # do the basic sort.
        # 'sortorder' is a tuple
        if not skipsort and not search_sort:
            issues = self._dosort(issues, sortorder)
            
        if reverse or rget('sortorder','')=='ignore':
            issues.reverse()

        # containers is not order, just filter it
        has_managerrole = self.hasManagerRole()
        for issue in issues:
            # if confidential and (if) Manager show, if not skip
            if not issue.IsConfidential() or has_managerrole:
                if filter == {}:
                    rlist.append(issue)
                else:
                    passsections = 1
                    for section in issue.sections:
                        if ss(section) in filter['sections']:
                            passsections = 0
                            break
                    # pass the filter lists

                    if ss(issue.status) not in filter['status'] \
                       and ss(issue.urgency) not in filter['urgency'] \
                       and ss(issue.type) not in filter['type'] \
                       and passsections and \
                       self._filter_fromnameemail(issue, filter):
                        rlist.append(issue)

                
        return rlist

    def _pop_from_list(self, item, olist):
        """ remove one item from a list """
        nlist = []
        for oitem in olist:
            if item!=oitem:
                nlist.append(oitem)
        return nlist
    
    def _filter_fromnameemail(self, issue, filter):
        """ do the advanced filtering.
            If filter has ignorefromname then reject this issue.
            If filter has showfromname, then ignore this issue
            if not == showfromname
        """
        # if asked to be ingored return false
        if filter.has_key('ignorefromname') and filter['ignorename'] !='' and \
           ss(issue.fromname) == filter['ignorefromname']:
            return false
        elif filter.has_key('ignoreemail') and filter['ignoreemail'] !='' and \
           ss(issue.email) == filter['ignoreemail']:
            return false
        elif filter.has_key('showfromname') and filter['showfromname'] !='' and \
           ss(issue.fromname) != filter['showfromname']:
            return false
        elif filter.has_key('showemail') and filter['showemail'] !='' and \
           ss(issue.email) != filter['showemail']:
            return false
        else:
            # not trapped in anything
            return true
        
    def _dosort(self, seq, key):
        """ do the actual sort """
        if type(key) not in [TupleType, ListType]:
            key = (key,)
        return sequence.sort(seq, (key,))
    
    def getBatchStart(self):
        """ return the batchstart value """
        try:
            return int(self.REQUEST.get('start',0))
        except:
            return false

    def getBatchSize(self):
        """ return the batchsize value """
        request = self.REQUEST
        if request.has_key('show') and request.get('show')=='all':
            return 100
        try:
            return int(request.get('size', self.default_batch_size))
        except:
            return 0
        
        
        
    def getCatalog(self):
        """ return the installed ICatalog object """
        if hasattr(self, 'ICatalog'):
            return self.ICatalog
        else: # backward compatability
            return self.Catalog

    def doDispatchOnSubmit(self):
        """ Check if we shall dispatch emails out """
        return getattr(self, 'dispatch_on_submit', DEFAULT_DISPATCH_ON_SUBMIT)

    def doStopCache(self):
        """ return the stop_cache property """
        return getattr(self, 'stop_cache', DEFAULT_STOP_CACHE)

    def HighlightQ(self, text, q=None, highlight_html=None):
        """ Highlight a piece of a text from q """
        if highlight_html is None:
            highlight_html = '<span class="q_highlight">%s</span>'
        if q is None:
            # then look for it in REQUEST
            q = self.REQUEST.get('q',None)
        if q is None:
            return text
        else:
            transtab = string.maketrans('/ ','_ ')
            q=string.translate(q, transtab, '?&!;()<=>*#[]{}')
            mem={}
            positions = []
            last_position = 0
            n_text = ''
            for q in self.QasList(q):
                r=re.compile(q, re.IGNORECASE)
                for i in r.findall(text):
                    r= "__%s__"%self._getrandstr()
                    mem[r]=i
                    text = self._text_replace(text, i, r)
            for k,v in mem.items():
                text = text.replace(k,highlight_html%v)
            return text
        
    def _text_replace(self, text, old, new):
        """ A custom string replace that doesn't have choke on tags.
        Don't do string replace on tags basically."""
        t=[]
        for part in text.split('<'):
            if part.find('>')>-1:
                t.append('<%s>'%part[0:part.find('>')])
                t.append(part[part.find('>')+1:].replace(old, new))
            else:
                t.append(part.replace(old,new))
        return ''.join(t)        

    def _getrandstr(self,l=5):
        """ """
        pool="0123456789"
        s=''
        for i in range(l):
            s='%s%s'%(s,random.choice(list(pool)))
        return s


    def colorizeThreadChange(self, title):
        """ Make "Changed status from Open to...
            to "Changed status from <span style="color:red;">Open</span> to...
        """
        highlight_html = '<span style="color:#CC0000">%s</span>'
        for status in self.getStatuses():
            title = self.HighlightQ(title, status, highlight_html)
        return title

    
    def QasList(self, q):
        """ q is a string that might contain 'and' and/or 'or'.
        Remove that and make it a list. """
        r=re.compile(r"\band\b|\bor\b", re.IGNORECASE)
        return r.sub("", q).split()

    def StopCache(self):
        """ Maybe we should set some cachepreventing headers """
        if self.doStopCache():
            response = self.REQUEST.RESPONSE
            response.setHeader('Expires', DateTime().toZone('GMT').rfc822())
            response.setHeader('Last-Modified', DateTime().toZone('GMT').rfc822())
    
    def HeadingLinks(self, display, sortname, default=0, startreversed=0):
        """Returns a hyperlink that can
           be used for resorting the listing
        """
        request = self.REQUEST
        querystring = request.QUERY_STRING

        ascending = self.www['ascarrow.gif'].tag(hspace=2)
        descending = self.www['descarrow.gif'].tag(hspace=2)
        invisible = self.www['invisiblearrow.gif'].tag(hspace=2)
        
        
        if request.has_key('reverse') and (request['reverse'] == 1 or request['reverse'] == '1'):
            reverse = 1
        elif startreversed and not (request.has_key('sortorder') and request['sortorder']==sortname):
            reverse = 1
        else:
            reverse = 0
        if request.has_key('sortorder') and request['sortorder'] == sortname or (default and not request.has_key('sortorder')):
            # have sorted by this, just let them reverse
            if reverse:
                ps = {'sortorder':sortname, 'reverse':None}
                url = self.aurl(request.URL, ps)
                url = self._addQuerystring(url)
                return """%s<a class="headinglink_current" href="%s"
                title="%s %s">%s</a>"""\
                       % (descending, url, SORT_BY, display, display )
            else:
                ps = {'sortorder':sortname, 'reverse':1}
                url = self.aurl(request.URL, ps)
                url = self._addQuerystring(url)
                return """%s<a class="headinglink_current" href="%s"
                title="%s">%s</a>"""\
                       % (ascending, url, SORT_REVERSE, display )
        else:
            if startreversed:
                ps = {'sortorder':sortname, 'reverse':1}
                url = self.aurl(request.URL, ps)
                url = self._addQuerystring(url)
                return """%s<a class="headinglink" href="%s"
                title="%s %s">%s</a>""" \
                   % (invisible, url, SORT_BY, display, display )
            else:
                ps = {'sortorder':sortname, 'reverse':None}
                url = self.aurl(request.URL, ps)
                url = self._addQuerystring(url)
                return """%s<a class="headinglink" href="%s"
                title="%s %s">%s</a>""" \
                   % (invisible, url, SORT_BY, display, display )

    def _addQuerystring(self, url):
        """ Add REQUEST querystring """
        querystring = self.REQUEST.get('QUERY_STRING','')
        if querystring is not None and querystring.strip()!='':
            url = "%s?%s"%(url, querystring)
        return url
    

    # Recent history SearchTerm
    #

    def RememberSearchTerm(self, q, yield):
        """ Stick this in a session variable """
        request = self.REQUEST
        key = RECENTHISTORY_SEARCHKEY
        
        searches = self.get_session(key, [])
        as_dict = {'q':q,'yield':yield}
        request.set('NotYetRecent',as_dict)
        if as_dict not in searches:
            searches.append(as_dict)
            self.set_session(key, searches)


    def hasRecentSearchTerms(self):
        """ check if any exists """
        request = self.REQUEST
        session = request.SESSION
        key = RECENTHISTORY_SEARCHKEY
        return self.get_session(key, {})!={}

    
    def getRecentSearchTerms(self):
        """ Return if any exists """
        request = self.REQUEST
        key = RECENTHISTORY_SEARCHKEY

        searches = self.get_session(key, {})
        searches.reverse()
        return searches


    # Recent history IssueVisit
    #


    def RememberIssueVisit(self, issueid):
        """ Remember that this issue has been visited """
        request = self.REQUEST

        key = RECENTHISTORY_ISSUEIDVISITKEY
        
        if type(issueid)!=StringType:
            # we only want objects id
            issueid = issueid.getId()

        visits = self.get_session(key, [])
        added_issueids = self.getRecentAddedIssues(ids=1)
        if issueid not in visits and issueid not in added_issueids:
            visits.append(issueid)
            self.set_session(key, visits)
            request.set('NotYetRecent',issueid)


    def hasRecentIssueVisits(self):
        """ check if any exists """
        if self.getRecentIssueVisits():
            return true
        else:
            return false
            
    def getRecentIssueVisits(self):
        """ Return if any exists """
        request = self.REQUEST

        key = RECENTHISTORY_ISSUEIDVISITKEY
        issueids = self.get_session(key, [])
        # make them objects
        issues=[]
        for issueid in self.filterTooRecent(issueids):
            try:
                issues.append(getattr(self, issueid))
            except:
                # Could have been deleted
                pass
        issues.reverse()
        return issues


    # Recent history AddedIssue
    #


    def RememberAddedIssue(self, issueid):
        """ Stick this in a session variable """
        request = self.REQUEST

        key = RECENTHISTORY_ADDISSUEIDKEY

        if type(issueid)!=StringType:
            # we only want objects id
            issueid = issueid.getId()
        added = self.get_session(key, [])
        if issueid not in added:
            added.append(issueid)
            self.set_session(key, added)
            request.set('NotYetRecent',issueid)

    def hasRecentAddedIssues(self):
        """ check if any exists """
        if self.getRecentAddedIssues():
            return true
        else:
            return false
            
    def getRecentAddedIssues(self, ids=0):
        """ Return if any exists """
        request = self.REQUEST

        key = RECENTHISTORY_ADDISSUEIDKEY

        issueids = self.get_session(key, [])
        # make them objects
        if ids:
            return issueids
        issues=[]
        for issueid in self.filterTooRecent(issueids):
            try:
                issues.append(getattr(self, issueid))
            except:
                # Could have been deleted
                pass
        issues.reverse()
        return issues


    def hasRecentHistory(self):
        """ check if anything is stored """
        test1 = self.hasRecentIssueVisits()
        test2 = self.hasRecentSearchTerms()
        test3 = self.hasRecentAddedIssues()
        return test1 or test2 or test3


    def filterTooRecent(self, recenthistory):
        """ Go through list and take out something too new """
        request = self.REQUEST
        too_recent_element = request.get('NotYetRecent',-1)
        n_recenthistory = []
        for each in recenthistory:
            if each != too_recent_element:
                n_recenthistory.append(each)
        return n_recenthistory
        


    # Misc scripts
    #

    def defineInstanceSessionKey(self, key):
        """ We use the default session key, but add to it for this
        issuetracker only. """
        id = self.getRoot().getId()
        return '%s-%s'%(key, id)

    def defineInstanceCookieKey(self, key):
        """ We use the default cookie key, but add to it for this
        issuetracker only. """
        # since that method is the same
        return self.defineInstanceSessionKey(key)
    
    def PropertiesWizardRemember(self, **kw):
        """ remember things for the wizard in the SESSION """
        had_before = self.get_session('propertieswizard',{})
        for key, value in kw.items():
            had_before[key] = value
        self.set_session('propertieswizard', had_before)
        
    def MoveWizardSession2REQUEST(self):
        """ move the items from the SESSION to REQUEST """
        request = self.REQUEST
        propertieswizard = self.get_session('propertieswizard',{})
        for key, value in propertieswizard.items():
            request.set(key, value)
            

    # POP3 
    #

    def getPOP3Accounts(self):
        """ return the POP3 Account objects """
        root = self.getPOP3Root()
        return root.objectValues(POP3ACCOUNT_METATYPE)
    
    security.declareProtected('View management screens','createPOP3Account')
    def createPOP3Account(self, hostname, username,
                          password, confirm=None, REQUEST=None):
        """ create POP3Account object """
        genid = "%s-%s"%(hostname, username)
        genid = genid.lower().strip()
        genid = IssueTrackerUtils.safeId(genid, nospaces=1)

        if confirm is not None:
            if password.strip() != confirm.strip():
                raise "PasswordMissmatch", "Password and confirm does not match"
            
        root = self.getPOP3Root()
        if hasattr(root, genid):
            raise "DuplicateId", "POP3Account already exists"
        
        pop3account = IssueTrackerEmail.POP3Account(genid, hostname, 
                                                    username, password)
        
        root._setObject(genid, pop3account)
        pop3account = getattr(root, genid)
        
        if REQUEST is not None:
            url = self.getRootURL()+'/manage_POP3ManagementForm'
            url = '%s?manage_tabs_message=%s'%(url, 'POP3 Account created')
            response = self.REQUEST.RESPONSE
            response.redirect(url)
        
    security.declareProtected('View management screens','editPOP3Account')
    def editPOP3Account(self, id, hostname=None, username=None, 
                        password=None, confirm=None,
                        password_dummy=None, REQUEST=None):
        """ edit POP3 account details """
        account = self.getPOP3Account(id)
        if hostname is not None and hostname.strip() != '':
            account.editAccount(hostname=hostname.strip())
        if username is not None and username.strip() != '':
            account.editAccount(username=username.strip())
        if password is not None and password.strip() != password_dummy:
            if password.strip() == confirm.strip():
                account.editAccount(password=password.strip())
            else:
                raise "PasswordMissmatch", "Password and confirm does not match"
            
        if REQUEST is not None:
            # redirect back to POP3 management form
            url = self.getRootURL()+'/manage_POP3ManagementForm'
            url = '%s?pop3accountid=%s'%(url, self.getId())
            url = '%s&manage_tabs_message=%s'%(url, 'POP3 Account saved')
            response = self.REQUEST.RESPONSE
            response.redirect(url)
                    
    security.declareProtected('View management screens','createAcceptingEmail')
    def createAcceptingEmail(self, id, email_address, defaultsections=None, 
                             default_type=None, default_urgency=None,
                             send_confirm=0, REQUEST=None):
        """ create accepting email objet in this account """
        account = self.getPOP3Account(id)
        if defaultsections is None:
            defaultsetions = self.defaultsections
        if default_type is None:
            default_type = self.default_type
        if default_urgency is None:
            default_urgency = self.default_urgency
            
        email_address = email_address.strip()
        if not self.ValidEmailAddress(email_address):
            raise "InvalidEmailAddress", "Email address is invalid"

        always_notify = ",".join(self.always_notify)
        always_notify = self.preParseEmailString(always_notify,aslist=1)
        if email_address.lower() in [x.lower() for x in always_notify]:
            raise "BusyEmailAddress", "%s is already used as always-notify"%\
                  email_address
        
        genid = email_address.replace('@','-at-').lower()
        account.createAcceptingEmail(genid, email_address, defaultsections,
                                     default_type, default_urgency,
                                     send_confirm)
        
        if REQUEST is not None:
            url = self.getRootURL()+'/manage_POP3ManagementForm'
            url = '%s?pop3accountid=%s'%(url, id)
            url = '%s&manage_tabs_message=%s'%(url, 'Accepting email created')
            response = self.REQUEST.RESPONSE
            response.redirect(url)
        
        
    def hasAcceptingEmails(self, id):
        """ return if any accepting emails """
        return len(self.getAcceptingEmails(id))>0
    
    def getAcceptingEmails(self, id):
        """ return accepting email objects """
        if getattr(id, 'meta_type','') == POP3ACCOUNT_METATYPE:
            root = id
        else:
            root = self.getPOP3Account(id)
        return root.objectValues(ACCEPTINGEMAIL_METATYPE)
    
    def getPOP3Account(self, id):
        """ get an object by id """
        return getattr(self.getPOP3Root(), id)
    

    security.declareProtected('View management screens','saveAcceptingEmails')
    def saveAcceptingEmails(self, id, allids):
        """ save all accepting emails. Find info via REQUEST object """
        request = self.REQUEST
        #return request
        account = self.getPOP3Account(id)
        for each_id in allids:
            acceptingemail = getattr(account, each_id)
            rkey_email_address = 'email_address-%s'%each_id
            rkey_defaultsections = 'defaultsections-%s'%each_id
            rkey_default_type = 'default_type-%s'%each_id
            rkey_defaul_urgency = 'default_urgency-%s'%each_id
            rkey_send_confirm = 'send_confirm-%s'%each_id

            email_address = request.get(rkey_email_address)
            if not self.ValidEmailAddress(email_address):
                raise "InvalidEmail", "Invalid email address %s"%email_address
            defaultsections = request.get(rkey_defaultsections)
            default_type = request.get(rkey_default_type)
            default_urgency = request.get(rkey_defaul_urgency)
            send_confirm = request.get(rkey_send_confirm, 0)
            
            acceptingemail.editDetails(email_address, defaultsections,
                                       default_type, default_urgency,
                                       send_confirm)
            
        url = self.getRootURL()+'/manage_POP3ManagementForm'
        url = '%s?pop3accountid=%s'%(url, id)
        url = '%s&manage_tabs_message=Accepting emails saved'%url
        response = request.RESPONSE
        response.redirect(url)
    
    
        
    security.declareProtected('View management screens','manage_delPOP3Accounts')
    def manage_delPOP3Accounts(self, ids=[], REQUEST=None):
        """ delete some POP3 Accounts """
        if type(ids) == type('s'):
            ids = [ids]
            
        root = self.getPOP3Root()
        root.manage_delObjects(ids)
        if REQUEST is not None:
            if len(ids)==0:
                mtm = "Nothing to delete"
            else:
                mtm = "Deleted %s POP3 Accounts"%len(ids)
            page = self.manage_POP3ManagementForm
            return page(self.REQUEST, manage_tabs_message=mtm)
        
    def getPOP3Root(self):
        """ return root/pop3 folder object. Create if necessary """
        return getattr(self.getRoot(), 'pop3')
    
    def manage_delAcceptingEmails(self, id, ids=[], REQUEST=None):
        """ delete some accepting email objects """
        account = self.getPOP3Account(id)
        if type(ids) == type('s'):
            ids = [ids]
        account.manage_delObjects(ids)
        
        if REQUEST is not None:
            url = self.getRootURL()+'/manage_POP3ManagementForm'
            url = '%s?pop3accountid=%s'%(url, id)
            url = '%s&manage_tabs_message=%s accepting emails deleted'%\
                (url, len(ids))
            response = self.REQUEST.RESPONSE
            response.redirect(url)
            
    def check4MailIssues(self):
        """ connect to a pop3 account and possibly create 
        some issues """
        
        if email_Parser is None:
            raise "NoEmailPackage", "The email package is not installed"
        
        allemails = []
        for account in self.getPOP3Accounts():
            M = poplib.POP3(account.getHostname())
            M.user(account.getUsername())
            M.pass_(account._password)
            
            # Get messages parsed
            #
            emails = self.getPOP3Messages(M, account)

            # Close the POP3 connection
            #
            M.quit()
            
            # Now we want to extract possible issue data
            #
            emails = self._appendEmailIssueData(emails, account)
            allemails.extend(emails)
            
        # In case we only have an email address
        combos = self.getEmailFromnameCombos()
        
        # Now, create the issues
        #
        count = 0
        for e in allemails:
            if e.get('fromname','') == '':
                e['fromname'] = combos.get(e['email'].lower(),'')
                e['fromname'] = e['fromname'].replace('<','').replace('>','')
                e['fromname'] = e['fromname'].strip()
            if e.get('IssueID') is not None:
                pass
            else:
                if e.has_key('display_format'):
                    display_format = e['display_format']
                else:
                    display_format = self.default_display_format
                create = self.createIssueObject
                object = create(None, e['title'], self.getStatuses()[0],
                                e['type'], e['urgency'], e['sections'],
                                e['fromname'], e['email'], '', 0, 0, 
                                e['body'], display_format,
                                e['date'], 1)
                count = count + 1
                for name, file in e.get('fileattachments', {}).items():
                    object.manage_addFile(name, file)
                
                # Possibly send a return email
                if e['acceptingemail'].doSendConfirm():
                    if hasattr(self, 'SendInboundEmailConfirm_script'):
                        script = self.SendInboundEmailConfirm_script
                        if e['fromname'] is not None and e['fromname'].strip() !='':
                            fromname = e['fromname']
                        else:
                            fromname = None
                        result = script(object, e['email'], fromname)
                    else:
                        LOG(self.__class__.__name__, WARNING, 
                            "No SendInboundEmailConfirm_script. See Deploy standards")

                    
                # Notify always notifyables
                self.sendAlwaysNotify(object.getId())
                        
        return "Created %s issues"%count

    def getEmailFromnameCombos(self):
        """ look through all issues and followups for combinations
        of fromname and email """
        combos = {}
        for issue in self.getRoot().objectValues(ISSUE_METATYPE):
            if not combos.has_key(issue.email.lower()):
                combos[issue.email.lower()] = issue.fromname
            for thread in issue.objectValues(ISSUETHREAD_METATYPE):
                if not combos.has_key(thread.email.lower()):
                    combos[thread.email.lower()] = thread.fromname
        return combos
    
    def _appendEmailIssueData(self, emails, account):
        """ inspect message for certain issue data. """
        allissueids = self.getRoot().objectIds(ISSUE_METATYPE)
        allsections = self.getSectionOptions()
        alltypes = self.types
        allurgencies = self.urgencies
        
        reg_issueids = "|".join(allissueids)
        reg_sections = "|".join(allsections)
        reg_types = "|".join(alltypes)
        reg_urgencies = "|".join(allurgencies)
        reg_structuredtext = r'STX|structuredtext'
        
        ign_case = re.IGNORECASE
        reg_issueids = re.compile(reg_issueids, ign_case)
        reg_sections = re.compile(reg_sections, ign_case)
        reg_types = re.compile(reg_types, ign_case)
        reg_urgencies = re.compile(reg_urgencies, ign_case)
        reg_structuredtext = re.compile(reg_structuredtext, ign_case)
        
        correct_caser = self._getCorrectCase
        for email in emails:
            s = email['subject'].strip()
            if s == '':
                self.sendReturnErrorEmail(email, "Subject line can not be empty")
                continue
            subject, parsable, delimiter = self._getParsableSubject(s)
            parsable = [x.strip() for x in parsable.split(',')]
            
            # look for issueid, i.e. a followup
            #issueids = []
            #for eachpart in parsable:
                #issueids.extend(reg_issueids.findall(eachpart))
            #if issueids:
                ## if any, accept only the first
                #email['issueid'] = correct_caser(issueids[0], allissueids)
                #parsable.remove(issueids[0])

            # Sections
            sections = []
            for eachpart in parsable:
                sections.extend(reg_sections.findall(eachpart))
            if sections:
                correct_sections = []
                for each in sections:
                    correct_sections.append(correct_caser(each, allsections))
                    try:
                        parsable.remove(each)
                    except:
                        pass
                email['sections'] = correct_sections
                
            # Type
            types = []
            for eachpart in parsable:
                types.extend(reg_types.findall(eachpart))
            if types:
                email['type'] = correct_caser(types[0], alltypes)
                parsable.remove(types[0])
                
            # Urgency
            urgencies = []
            for eachpart in parsable:
                urgencies.extend(reg_urgencies.findall(eachpart))
            if urgencies:
                email['urgency'] = correct_caser(urgencies[0], allurgencies)
                parsable.remove(urgencies[0])
                
            # Structured or plain text
            # This one is a bit special.
            structured_text = []
            for eachpart in parsable:
                structured_text.extend(reg_structuredtext.findall(eachpart))
            if structured_text:
                email['display_format'] = 'structuredtext'
                parsable.remove(structured_text[0])
                
            if parsable:
                leftovers = ', '.join(parsable)
                if delimiter in ['[',']']:
                    subject = "[%s] %s"%(leftovers, subject)
                elif delimiter == ':':
                    subject = "%s: %s"%(leftovers, subject)

            email['title'] = subject
            
            # Retrospect, and fill in with default values.
            # This is where we use the default* values from the
            # matching accepting email object
            acceptingemail = account.getAcceptingEmailbyTo(email['to'])
            email['acceptingemail'] = acceptingemail
            if not email.has_key('sections'):
                email['sections'] = acceptingemail.defaultsections
            if not email.has_key('type'):
                email['type'] = acceptingemail.default_type
            if not email.has_key('urgency'):
                email['urgency'] = acceptingemail.default_urgency
                

            extractor = self.preParseEmailString
            email['email'] = extractor(email['from'], allnotifyables=0)
            f = email['from'].replace(email['email'],'').strip()
            f = f.replace('<','').replace('>','').strip()
            email['fromname'] = f
            
            
        return emails
    
    def sendReturnErrorEmail(self, email, msg):
        """ Send a simple email when there is an error """
        mFrom = self.sitemaster_email
        mTo = email['from']
        mSubject = "[Autoreply] Inbound issue email incorrect"
        mBody = "There was an error in your inbound email to %s\n\n"%\
              self.getRoot().getTitle()
        mBody = mBody + "Error: %s"%msg
        body = '\r\b'.join(["From: %s"%mFrom,"To: %s"%mTo,
                            "Subject: %s"%mSubject, "",mBody])
        self.sendEmail(body, mTo, mFrom, mSubject)
        
    def _getParsableSubject(self, subject):
        """ check the subject line for what is really the parsable
        bit and the textual bit.
        Return (textual, parsable, delimiter) 
        Where delimiter is either '[' or ':'"""
        
        # default
        textual = subject
        delimiter = parsable = ''
        if subject[0]=='[' and subject[1:].find(']') > -1 and not \
           subject[-1]==']':
            # Used like this -  [Bug, Help] Bla bla bla
            parsable = subject[1:subject.find(']')].strip()
            textual = subject[subject.find(']')+1:].strip()
            delimiter = '['
        elif subject.find(':')>0:
            # Used like this -  Bug, Help: Bla bla bla
            parsable = subject[:subject.find(':')].strip()
            textual = subject[subject.find(':')+1:].strip()
            delimiter = ':'
        return textual, parsable, delimiter
    
        
    def _getCorrectCase(self, item, list):
        """ item might be 'abc' and list ['Abc','Def'] 
        then return 'Abc' """
        for correct_item in list:
            if item.lower()==correct_item.lower():
                return correct_item
        else:
            return item
            
            
    def getPOP3Messages(self, pop3instance, account, dele=1):
        """ get messages from pop3 object """
        numMessages = len(pop3instance.list()[1])
        emails = []
        accept_email_addresses = []
        for acceptingemail in account.getAcceptingEmails():
            accept_email_addresses.append(acceptingemail.email_address)
            
        basepath = os.path.join(INSTANCE_HOME, 'var') 
        for i in range(numMessages):
            tmpfilename = os.path.join(basepath, 'pop3%s.email'%i)
            
            f_w = open(tmpfilename, 'wb')
            for line in pop3instance.retr(i+1)[1]:
                f_w.write(line+'\n')
            f_w.close()
            
            p = email_Parser.Parser()
            f_r = open(tmpfilename, 'rb')
            msg = p.parse(f_r)
            f_r.close()
            os.remove(tmpfilename)
            
            e = {}
            e['fileattachments']={}
            for key, value in msg.items():
                e[key.lower()] = value
            for part in msg.walk():
                if part.get_main_type() == 'multipart':
                    continue
                name = part.get_param('name')
                if name is None:
                    name = part.get_filename()
                try:
                    content = part.get_payload(decode=1)
                except:
                    # This might happen if the part is too unnormal
                    # for the email package to deal with. In that
                    # case, this attachment is ignorable. Tough!
                    continue
                if name == None:
                    if self._isHTMLBody(content):
                        if html2safehtml is not None:
                            content = self._stripHTMLBody(content)
                            e['display_format'] = 'html'
                        else:
                            LOG(self.__class__.__name__, WARNING,
                                "stripogram module not installed to strip HTML emails")
                    e['body'] = content
                else:
                    e['fileattachments'][name] = content
            
            # Maybe it wasn't sent directly To, but CC
            if e.get('cc','') != '':
                e['to'] = "%s, %s"%(e.get('to',''), e['cc'])
            # check whom it's to
            extractor = self.preParseEmailString
            tolist = extractor(e['to'], aslist=1, allnotifyables=0)
            intersection = self._getIntersection(tolist, accept_email_addresses)
            if intersection:
                e['to'] = intersection[0]
                if dele:
                    pop3instance.dele(i+1)
                emails.append(e)
                
        return emails
                
    def _getIntersection(self, list1, list2):
        """ if 'A, C, D' in ['a','b'] should return true """
        intersection = []
        if list1 is None:
            return []
        elif not self.same_type(list1, []):
            list1 = [list1]
            
        if not self.same_type(list2, []):
            list2 = [list2]
        list2lower = [x.lower().strip() for x in list2]
        for item in list1:
            if item.lower().strip() in list2lower:
                intersection.append(item)
        return intersection
    
    def _isHTMLBody(self, body):
        """ check if the body is html encoded """
        if body is None:
            return false
        body = body.lower().strip()
        return body.startswith('<html>') and body.endswith('</html>')
        
    def _stripHTMLBody(self, body):
        """ strip out all HTML if possible from the email """
        accept_tags = ('b','strong','br','i','em','p','a',
                       'ol','ul','li')
        return html2safehtml(body, valid_tags=accept_tags)

        
    # Statistics
    #
    
    def totalCountStatus(self, statuslist):
        """ in a status list [['open',4], ...] 
        sum up all the numbers """
        count = 0
        for item in statuslist:
            count += item[1]
        return count
            
    def CountSections(self):
        """ for every section, count how many for each status 
        return as [['General', {'open':4, 'taken':6, ...}], ...] """
        res = []
        allsections = {}
        allissues = self.getRoot().objectValues(ISSUE_METATYPE)
        for issue in allissues:
            status = issue.status.lower()
            for section in issue.sections:
                if not allsections.has_key(section):
                    allsections[section] = {}
                if not allsections[section].has_key(status):
                    allsections[section][status] = 0
                allsections[section][status] += 1
                
        # add all zeros
        for section in self.getSectionOptions():
            if not allsections.has_key(section):
                allsections[section] = {}
            allsections[section] = self._allStatuses(allsections[section])
            res.append([section, allsections[section]])
                
        return res
                
    def _allStatuses(self, dict):
        """ dict might be {'open':2, 'taken':0}
        then make sure it as all possible statuses """
        for status in self.getStatuses():
            if not dict.has_key(status.lower()):
                dict[status] = 0
        return dict
    
    def totalCountSections(self, sectiondict):
        """ sum the total in {'open':2, 'taken':1, ...} """
        count = 0
        for value in sectiondict.values():
            count += value
        return count
    
    def issueInflux(self, from_date=None, till_date=None,
                    issues=None, returncount=0):
        """ calculate for different day periods approximately how 
        many issues are coming in """
        if from_date is not None and self.same_type(from_date, 's'):
            from_date = DateTime(from_date)
            
        if issues is not None:
            allissues = issues
        else:
            allissues = self.getRoot().objectValues(ISSUE_METATYPE)
            allissues = sequence.sort(allissues, (('issuedate',),))
            
        # if from_date is None, then make the first issue the from_date
        if from_date is None:
            from_date = allissues[0].issuedate
        if till_date is None:
            till_date = DateTime()

        count = 0 
        for issue in allissues:
            if issue.issuedate >= from_date and issue.issuedate < till_date:
                count += 1
        
        day_span = till_date - from_date
        
        if returncount:
            return count
        else:
            issue_per_day = count / day_span
            return issue_per_day
            
        
    def issueInfluxbyPeriod(self, period=14):
        """ prepare a issues per period list """
        allissues = self.getRoot().objectValues(ISSUE_METATYPE)
        allissues = sequence.sort(allissues, (('issuedate',),))
        
        try:
            period = int(period)
        except:
            raise "PeriodNotInteger", "The period must be an integer"
        start_date = allissues[0].issuedate
        end_date = allissues[-1].issuedate
        difference_days = end_date - start_date
        
        influxes = []
        today = DateTime()
        highest = 0
        for i in range(int(difference_days/period)+1):
            from_date = start_date + i * period
            till_date = from_date + period
            if till_date > today:
                till_date = today
            data = {'from':from_date, 'till':till_date}
            influx = self.issueInflux(from_date, till_date,
                                      allissues, 1)
            data['influx'] = influx
            if influx > highest:
                highest = influx
            influxes.append(data)
            
        return influxes, highest
            
            
        
    
    # Templates
    #
    
    #def _hasTemplate(self, id):
        #""" check if a particular object id exists without
        #acquisition """
        #root = self.getRoot()
        #base = getattr(root, 'aq_base', root)
        #return hasattr(base, id)
    
    
    #def AddIssue(self, *arg, **kw):
        #""" possibly soft-method-override """
        #if self._hasTemplate('AddIssue.zpt'):
            #return getattr(self, 'AddIssue.zpt')(mapping=kw)
        #else:
            #ptf = PTF('zpt/AddIssue', globals(), mapping=kw)
            #return ptf.__of__(self)()

    AddIssue = PTF('zpt/AddIssue', globals())
    QuickAddIssue = PTF('zpt/QuickAddIssue', globals())
    SearchForm = PTF('zpt/SearchForm', globals())
    user_searchform = PTF('zpt/user_searchform', globals())
    submitform = PTF('zpt/submitform', globals())
    main_tabs = PTF('zpt/main_tabs', globals())
    tell_a_friend = PTF('zpt/tell_a_friend', globals())
    subscription = PTF('zpt/subscription', globals())
    recent_history_view = PTF('zpt/recent_history_view', globals())
    quickpreviewform = PTF('zpt/quickpreviewform', globals())
    previewform = PTF('zpt/previewform', globals())
    latestIssues = PTF('zpt/latestIssues', globals())
    list_issues_top_bar = PTF('zpt/list_issues_top_bar', globals())
    filter_options = PTF('zpt/filter_options', globals())
    security.declareProtected('View', 'ListIssues')
    CompleteList = PTF('zpt/CompleteList', globals(), __name__='CompleteList')
    security.declareProtected('View', 'CompleteList')
    ListIssues = PTF('zpt/ListIssues', globals(), __name__='ListIssues')
    security.declareProtected('View','index_html')
    index_html = PTF('zpt/index_html', globals())
    Statistics = PTF('zpt/Statistics', globals())
    ShowIssue = PTF('zpt/ShowIssue', globals())
    ShowIssueData = PTF('zpt/ShowIssueData', globals())
    ShowIssueThreads = PTF('zpt/ShowIssueThreads', globals())


IssueTracker.__dict__['About.html'] = IssueTracker.About
IssueTracker.__dict__['rss-0.91.xml'] = IssueTracker.RSS091
# default RSS
IssueTracker.__dict__['rss.xml'] = IssueTracker.RSS091

InitializeClass(IssueTracker)



###############   ISSUETRACKERISSUE    ###################

class IssueTrackerIssue(IssueTracker):
    """ Issues class as containers """

    meta_type = ISSUE_METATYPE
    icon = '%s/issue.gif'%ICON_LOCATION
    
    _properties=({'id':'title',         'type': 'string', 'mode':'w'},
                 {'id':'issuedate',     'type': 'date',   'mode':'w'},
                 {'id':'status',        'type': 'string', 'mode':'w'},
                 {'id':'type',          'type': 'string', 'mode':'w'},
                 {'id':'urgency',       'type': 'string', 'mode':'w'},
                 {'id':'sections',      'type': 'lines',  'mode':'w'},
                 {'id':'fromname',      'type': 'string', 'mode':'w'},
                 {'id':'email',         'type': 'string', 'mode':'w'},
                 {'id':'url2issue',     'type': 'string', 'mode':'w'},
                 {'id':'confidential',  'type': 'boolean','mode':'w'},
                 {'id':'hide_me',       'type': 'boolean','mode':'w'},
                 {'id':'description',   'type': 'text',   'mode':'w'},
                 {'id':'display_format','type': 'string', 'mode':'w'},
                 {'id':'subscribers',   'type': 'lines',  'mode':'w'},                 
                 )
    
    security=ClassSecurityInfo()

    manage_options = (
        {'label':'Contents', 'action':'manage_main'},
        {'label':'Properties', 'action':'manage_propertiesForm'}
        )

    def __init__(self, id, title, status, issuetype, urgency, sections,
                 fromname, email, url2issue, confidential, hide_me,
                 description, display_format, issuedate='', subscribers=[]):
        """ init an Issue object """
        self.id = str(id)
        self.title = title.strip()
        if type(issuedate) == StringType:
            issuedate = IssueTrackerUtils.DateStr2DateDate(issuedate)
        self.issuedate = issuedate
        self.status = status
        self.type = issuetype
        self.urgency = urgency
        self.sections = sections
        self.fromname = fromname
        self.email = email
        self.url2issue = url2issue
        self.confidential = confidential
        self.hide_me = hide_me
        self.description = description
        if display_format:
            self.display_format = display_format
        else:
            self.display_format = self.default_display_format
        self.subscribers = subscribers
        
    def assertAllProperties(self):
        """ make sure issue has all properties """
        props = {'title':'', 'issuedate':DateTime(), 
                 'status':self.getStatuses()[0], 'type':self.default_type,
                 'urgency':self.default_urgency,
                 'sections':self.defaultsections,
                 'fromname':'', 'email':'', 'url2issue':'',
                 'confidential':false, 'hide_me':false,
                 'description':'', 
                 'display_format':self.default_display_format,
                 'subscribers':[]}
                 
        count = 0
        for key, default in props.items():
            if not self.__dict__.has_key(key):
                self.__dict__[key] = default
                count += 1
        return count
                
    security.declareProtected('View', 'index_html')
    def index_html(self, REQUEST=None):
        """ show the issue """
        if not self.confidential or self.hasManagerRole():
            return self.ShowIssue()
        else:
            response = self.REQUEST.RESPONSE
            listpage = '/%s'%self.whichList()
            response.redirect(self.getRootURL()+listpage)

    def IsConfidential(self):
        """ check confidential property """
        return getattr(self, 'confidential', false)
    
        #if hasattr(self, 'confidential') and self.confidential:
            #return true
        #else:
            #return false
        
    def ModifyIssue(self, action='Addfollowup'):
        """ advanced change to issue properties """
        request = self.REQUEST

        issueobject = self
        action = action.lower().strip()
        oldstatus = self.status
        prefix = self.issueprefix
                
        if not request.has_key('comment_description'):
            request.set('comment_description','')

        if not request.has_key('display_format'):
            request.set('display_format',self.default_display_format)

        req_manager = 1 # require manager by default
        dummychange = 0
        addfollowup = 0
        past_tense=None
        statuses_and_verbs = self.getStatusesMerged(asdict=1)
        lowercase_values = {}
        statuses_and_verbs_reversed = {}
        for key, value in statuses_and_verbs.items():
            lowercase_values[value.lower()] = value
            statuses_and_verbs_reversed[value] = key
            
        if lowercase_values.has_key(action):
            past_tense = statuses_and_verbs_reversed[lowercase_values[action]]
            addfollowup = 1
        
        #if action == 'open':
            #past_tense='open'
            #addfollowup = 1
        #if action == 'take':
            #past_tense='taken'
            #addfollowup = 1
        #if action == 'put on hold':
            #past_tense='on hold'
            #addfollowup = 1
        #if action == 'reject':
            #past_tense='rejected'
            #addfollowup = 1
        #if action == 'complete':
            #past_tense='completed'
            #addfollowup = 1
        elif action == 'delete':
            DeleteIssue(self, self.id)
            addfollowup = 0
            redirect_url = request.URL2
        elif action == 'addfollowup':
            addfollowup = 1
            dummychange = 1
            req_manager = 0

        if req_manager and not self.hasManagerRole():
            # the chosen action requires manager role,
            # but the person is not a manager
            self.login(self,request)

        if dummychange:
            # Still should trigger a new bobobase_modification_time
            status = self.status
            
        if past_tense is not None:
            self.status = past_tense
        else:
            # anything else means that the comment_description
            # cannot be empty
            if request['comment_description'].strip()=='':
                addfollowup=0
                err = """When submitting a followup you can not leave the description empty"""
                request.set('previewissue',1)
                request.set('SubmitError', [err])
                return self.ManagerOptionsExtend()
                #return "<ul><li>%s</li></ul>"%err
                #redirect_url = request.URL1

        # most actions may perhaps add a little comment
        if addfollowup:

            genid = issueobject.generateID(4, 'issuethread'+prefix,
                                           meta_type=ISSUETHREAD_METATYPE)
            
            if action == 'addfollowup':
                gentitle = 'Added Issue followup'
            else:
                gentitle = 'Changed status from %s to %s'%\
                           (oldstatus.capitalize(), past_tense.capitalize())

            # fix variables
            fromname = request.get('fromname','')
            email = request.get('email','')
            display_format = request.get('display_format',
                                         self.default_display_format)
                
            # remember visitor
            self.set_cookie(self.getCookiekey('name'),fromname)
            self.set_cookie(self.getCookiekey('email'),email)
            self.set_cookie(self.getCookiekey('display_format'),display_format)

            # update Thread object
            title  = gentitle
            comment = request.comment_description
            threaddate = getNowdate()

            # create an Thread object
            create_method = issueobject._createThreadObject
            followupobject = create_method(genid, title, comment,
                                           threaddate, fromname,
                                           email, display_format)
            followupobject.index_object(idxs=['id','comment'])

            # Also upload the fileattachments
            self._moveTempfiles(followupobject)
    
            # upload new file attachments
            self._uploadFileattachments(followupobject)
            
            self.nullifyTempfolderREQUEST()

            if request.has_key('notify'):
                        
                # now, create and email-alert-queue object
                # using filtered email address
                # get who to notify        
                email_addresses = self.Others2Notify(issueobject.getId(),
                                                     do='email',emailtoskip=email)

                if len(email_addresses) > 0:

                    # create id for notification
                    notifyid = self.generateID(4, prefix+"notification",
                                               meta_type='Issue Notification')

                    title = self.title
                    change = gentitle
                    issueID = self.id
                    anchorname = len(self.objectIds(ISSUETHREAD_METATYPE))
                    emails = email_addresses
                    date = getNowdate()
                    if not hasattr(self.getRoot(), issueID):
                        raise "IssueNonexisting", "This notification has no issue"
                    else:
                        notification = IssueTrackerNotification(
                                        notifyid, title, change, issueID,
                                        comment, emails, anchorname, fromname,
                                        date)
                        self._setObject(notifyid, notification)
                        notifyobject = getattr(self, notifyid)
                      
                        # use the dispatcher to try to send
                        # this notification right now.
                        request.set('loop_over',[notifyobject])
                        # there is no big deal if the dispatcher crashes here
                        # because the notification is saved and the dispatcher
                        # can be invoked some other time manually
                        if self.doDispatchOnSubmit():
                            try:
                                getattr(self,'dispatcher.html')(self, request)
                            except:
                                LOG(self.meta_type, PROBLEM,
                                    'Email could not be sent',
                                    error=sys.exc_info())
            
            
            redirect_url = '%s#i%s'%(issueobject.absolute_url(),
                                     len(issueobject.objectIds(ISSUETHREAD_METATYPE)))
        # catalog
        issueobject.reindex_object()
        # ready ! redirect!
        request.RESPONSE.redirect(redirect_url)

    def _createThreadObject(self, id, title, comment, threaddate,
                            fromname, email, display_format):
        """ Crudely create thread object. No checking. """
        thread = IssueTrackerIssueThread(id, title, comment, threaddate,
                                         fromname, email, display_format)
        self._setObject(id, thread)

        # get that object
        threadobject = getattr(self, id)

        return threadobject

    def change_issueAttributes(self, sections, issuetype, urgency, REQUEST=None):
        """ Custom edit of only these properties """
        if not type(sections)==type([]):
            sections=[sections]
        n_sections = []
        for section in sections:
            if section in self.getSectionOptions():
                n_sections.append(section)
        self.sections=n_sections
        
        if not type(issuetype)==type('s'):
            issuetype=str(issuetype)
        if issuetype in self.types:
            self.type = issuetype

        if not type(urgency)==type('s'):
            urgency=str(urgency)
        if urgency in self.urgencies:
            self.urgency = urgency
            
        if REQUEST is not None:
            REQUEST.RESPONSE.redirect(self.absolute_url())
    
    def Others2Notify(self, issueID, do='print', emailtoskip=None):
        """
        Returns a list of names and emails of guys to notify on this Issue. 
        Take the hide_me and confidential factor into consideration.
        
        if do == 'print':
            display the people on screen (using hide_me and confidential factor)
        if do == 'email':
            return a pure list of all email address
        """

        if emailtoskip is None:
            if self.has_cookie(self.getCookiekey('email')):
                emailtoskip = self.get_cookie(self.getCookiekey('email'))
            else:
                emailtoskip = '@@@@@'

        fromstrings = []
        emails = []

        issue = getattr(self.getRoot(), issueID)

        # proceed only if Manager or open-name on the issue
        if self.hasManagerRole() or not issue.hide_me:
            # self.email is the email of the thread
            # emailtoskip comes from the param/cookie if applicable
            if not issue.email == emailtoskip:
                if do =='email':
                    fromstring = getFromstring(issue.fromname, issue.email, 
                                               email_must=1, email_pure=1)
                else:
                    fromstring = getFromstring(issue.fromname, issue.email)
                    emails.append(issue.email.lower())
                # fromstring looks like this now: 
                # "Peter Bengtsson <mail@peterbe.com>"
                # or ['mail@peterbe.com'] depending on 'do'
                # ...maybe
                if fromstring and fromstring not in fromstrings:
                    fromstrings.append(fromstring)
                           
                        
        
        for thread in issue.objectValues(ISSUETHREAD_METATYPE):
            if not thread.email == emailtoskip:
                if do =='email':
                    fromstring = getFromstring(thread.fromname, thread.email,
                                               email_must=1, email_pure=1)
                else:
                    fromstring = getFromstring(thread.fromname, thread.email)
                    if thread.email.lower() not in emails:
                        emails.append(thread.email.lower())
                if fromstring and fromstring not in fromstrings and \
                   thread.email.lower() not in emails:
                    fromstrings.append(fromstring)
                    
        if do == 'email':
            for subscriber in self.getSubscribers():
                if subscriber not in fromstrings:
                    fromstrings.append(subscriber)
        else:
            notifyables = self.getNotifyablesEmailName()
            for subscriber in self.getSubscribers():
                if notifyables.has_key(subscriber):
                    fromstring = getFromstring(notifyables[subscriber],
                                               subscriber)
                else:
                    fromstring = subscriber
                if fromstring and fromstring not in fromstrings and \
                   str(fromstrings).find(fromstring) == -1:
                    fromstrings.append(fromstring)

        return fromstrings

    security.declareProtected('View', 'issue2TXT')
    def issue2TXT(self, RESPONSE):
        """ prepare an issue for download """
        if not self.confidential or self.hasManagerRole():

            if RESPONSE is not None:
                RESPONSE.setHeader('Content-type','application/data')
                RESPONSE.setHeader('Content-Disposition','inline;filename=%s.txt' % (self.id))

            return self.issue2TXT_script(self)
        else:
            response = self.REQUEST.RESPONSE
            listpage = '/%s'%self.whichList()
            response.redirect(self.getRootURL()+listpage)

    security.declareProtected('Access contents information', 'issue2TXT_script')
    def issue2TXT_script(self,issue):
        """ the textual representation of an issue """

        nl="\r\n"
        TXT =""
        today_fmt = self.ZopeTime().strftime(self.display_date)
        TXT += "Export of issue (%s) %s"%(issue.id, today_fmt)
        TXT += "%s--------------------------------%s%s"%(nl,nl,nl)
        TXT += " %s%s%s"%(issue.title,nl,nl)
        TXT += "Status:     %s%s"%(issue.status.capitalize(),nl)
        issuedate_fmt = issue.issuedate.strftime(self.display_date)
        TXT += "Issue date: %s%s"%(issuedate_fmt,nl)
        TXT += "Type:       %s%s"%(issue.type,nl)
        TXT += "Urgency:    %s%s"%(issue.urgency,nl)
        TXT += "Sections:   %s%s"%(string.join(issue.sections,', '),nl)
        TXT += "URL:        %s%s"%(issue.url2issue,nl*2)

        if issue.hide_me:
            TXT += "Name and email hidden%s"%nl
            if context.hasManagerRole():
                TXT += "Fromname:   %s%s"%(issue.fromname,nl)
                TXT += "Email:      %s%s"%(issue.email,nl*2)
        else:
            TXT += "Fromname:   %s%s"%(issue.fromname,nl)
            TXT += "Email:      %s%s"%(issue.email,nl*2)

        TXT += nl*2
        TXT += "%s%s"%(issue.description.replace('\n',nl),nl)

        threads = issue.objectValues(ISSUETHREAD_METATYPE)
        if len(threads)>0:
            TXT += "%s------------Followups------------%s"%(nl*2,nl*2)
            for thread in threads:
                TXT += "%s %s%s"%(nl,thread.title,nl*2)
                threaddate_fmt = thread.threaddate.strftime(self.display_date)
                TXT += "Followup date: %s%s"%(threaddate_fmt,nl)
                TXT += "Fromname:      %s%s"%(thread.fromname,nl)
                TXT += "Email:         %s%s"%(thread.email,nl*2)

                TXT += thread.comment.replace('\n',nl)
                TXT += nl*2

        return TXT

    def getOptionButtons(self):
        res=[]

        issuestatus = self.status.lower()
        for item in self.getStatusesMerged(aslist=1):
            status, verb = item
            if issuestatus != status.lower():
                action = 'IssueAction_%s'%verb.replace(' ','').capitalize()
                
                res.append([action, verb.capitalize()])
                
        #if self.status.lower() !='open':
            #res.append(['IssueAction_Open','Open'])
        #if self.status.lower() !='taken':
            #res.append(['IssueAction_Take','Take'])
        #if self.status.lower() != 'on hold':
            #res.append(['IssueAction_Onhold','Put on hold'])
        #if self.status.lower() != 'rejected':
            #res.append(['IssueAction_Rejected','Reject'])
        #if self.status.lower() != 'completed':
            #res.append(['IssueAction_Complete','Complete'])
        res.append(['IssueAction_Delete','Delete',])

        return res

    def ListThreads(self,issueID=None):
        """ Return the followup objects. The threads. """
        req = self.REQUEST
        rlist = []
        if issueID is not None:
            issueobject = getattr(self.getRoot(),issueID)
        else:
            issueobject = self
        threads = issueobject.objectValues(ISSUETHREAD_METATYPE)
        return threads
    
    def getLastThread(self, issueID=None):
        """ Return the last thread in an issue or None """
        allthreads = self.ListThreads(issueID)
        if allthreads:
            return allthreads[-1]
        else:
            return None
        
    def countFileattachments(self):
        """ return [no files in issue, no files in threads] """
        return len(self.ZopeFind(self, obj_metatypes=['File'], search_sub=1))
        
   # security.declareProtected('Manage ZCatalog Entries', 'manage_beforeDelete')
    def manage_beforeDelete(self, REQUEST=None, RESPONSE=None):
        """ uncatalog yourself """
        for thread in self.objectValues(ISSUETHREAD_METATYPE):
            thread.unindex_object()
        self.unindex_object()

    def index_object(self, idxs=[]):
        """A common method to allow Findables to index themselves."""
        path = '/'.join(self.getPhysicalPath())
        self.getCatalog().catalog_object(self, path, idxs=idxs)

    def unindex_object(self):
        
        """A common method to allow Findables to unindex themselves."""
        self.getCatalog().uncatalog_object('/'.join(self.getPhysicalPath()))
        
    def isSubscribing(self, email):
        """ check if in subscribers list """
        subscribers = self.getSubscribers()
        return str(email).strip().lower() in [x.lower() for x in subscribers]
    
    def getSubscribers(self):
        """ return subscribers list """
        return getattr(self, 'subscribers', [])
    
    def addSubscriber(self, name_or_email):
        """ add subscriber to subscribers list """
        subscribers = self.getSubscribers()
        if not self.ValidEmailAddress(name_or_email):
            for notifyable in self.getNotifyables():
                if name_or_email.strip() == notifyable.getName():
                    name_or_email = notifyable.getEmail()
                    break

        if not self.isSubscribing(name_or_email):
            subscribers.append(name_or_email)
            self.subscribers = subscribers
            
    def delSubscriber(self, email):
        """ remove item from subscribers list """
        n_subscribers = []
        for subscriber in self.getSubscribers():
            if email.lower().strip() != subscriber.lower().strip():
                n_subscribers.append(subscriber)
        self.subscribers = n_subscribers
            
    def Subscribe(self, subscribe_email, unsubscribe=0, REQUEST=None):
        """ analyse subscribe_email and add accordingly """
        emails = self.preParseEmailString(subscribe_email, aslist=1)
        for email in emails:
            if unsubscribe:
                self.delSubscriber(email)
            else:
                self.addSubscriber(email)
        
        if REQUEST is not None:
            REQUEST.RESPONSE.redirect(self.absolute_url())
            

    # Templates
    shownewIssueURL = PTF('zpt/shownewIssueURL', globals())
    manager_options = PTF('zpt/manager_options', globals())
    anonymous_options = PTF('zpt/anonymous_options', globals())
    form_followup = PTF('zpt/form_followup', globals())
    form_open = PTF('zpt/form_open', globals())
    form_take = PTF('zpt/form_take', globals())
    form_onhold = PTF('zpt/form_onhold', globals())
    form_reject = PTF('zpt/form_reject', globals())
    form_delete = PTF('zpt/form_delete', globals())
    form_complete = PTF('zpt/form_complete', globals())
    followup_preview = PTF('zpt/followup_preview', globals())


InitializeClass(IssueTrackerIssue)


############### ISSUETRACKERISSUETHREAD ##################


class IssueTrackerIssueThread(IssueTrackerIssue): 
    
    """ Issuethreads class """

    meta_type = ISSUETHREAD_METATYPE
    icon = '%s/issuethread.gif'%ICON_LOCATION
    
    #meta_types = []

    _properties=({'id':'title',         'type': 'string', 'mode':'w'},
                 {'id':'comment',       'type': 'text',   'mode':'w'},
                 {'id':'threaddate',    'type': 'date',   'mode':'w'},
                 {'id':'fromname',      'type': 'string', 'mode':'w'},
                 {'id':'email',         'type': 'string', 'mode':'w'},
                 {'id':'display_format','type': 'string', 'mode':'w'},
                 )
    
    security=ClassSecurityInfo()

    manage_options = (
        {'label':'Properties', 'action':'manage_propertiesForm'},
        {'label':'Contents', 'action':'manage_main'},                     
        )

    def __init__(self, id, title, comment, threaddate, fromname, email,
                 display_format=None, REQUEST=None):
        """ create thread """
        self.id = str(id)
        self.title = title
        self.comment = comment
        if type(threaddate) == StringType:
            threaddate = DateTime(threaddate)
        self.threaddate = threaddate
        self.fromname = fromname
        self.email = email
        if display_format:
            self.display_format = display_format
        else:
            self.display_format = self.default_display_format

    security.declareProtected('Manage ZCatalog Entries', 'manage_beforeDelete')
    def manage_beforeDelete(self, REQUEST=None, RESPONSE=None):
        """ uncatalog yourself """
        self.unindex_object()



InitializeClass(IssueTrackerIssueThread)




############### ISSUETRACKERNOTIFICATION ###################

class IssueTrackerNotification(SimpleItem.SimpleItem,
                               IssueTracker
                               ):
    """ Issue Tracker Notification """

    meta_type = NOTIFICATION_META_TYPE
    icon = '%s/notification.gif'%ICON_LOCATION

    security=ClassSecurityInfo()
    
    _properties=({'id':'title',       'type': 'string',  'mode':'w'},
                 {'id':'change',      'type': 'string',  'mode':'w'},
                 {'id':'issueID',     'type': 'string',  'mode':'w'},
                 {'id':'comment',     'type': 'text',    'mode':'w'},
                 {'id':'emails',      'type': 'lines',   'mode':'w'},
                 {'id':'anchorname',  'type': 'string',  'mode':'w'},
                 {'id':'fromname',    'type': 'string',  'mode':'w'},
                 {'id':'date',        'type': 'date',    'mode':'w'},
                 {'id':'dispatched',  'type': 'boolean', 'mode':'w'},
                 )

    manage_options = (
        {'label':'Properties', 'action':'manage_propertiesForm'},
        )

    def __init__(self, id, title, change, issueID, comment, emails,
                 anchorname, fromname, date=None, dispatched=0, REQUEST=None):
        """ create notification """
        issueID = str(issueID)
        self.id = str(id)
        self.title = title
        self.change = change
        self.issueID = issueID
        self.comment = comment
        self.emails = emails
        self.anchorname = anchorname
        self.fromname = fromname
        self.dispatched = dispatched
        if date is None:
            self.date = getNowdate()
        else:
            self.date = date

    def dispatch(self):
        """ dispatch self """
        self.REQUEST.set('loop_over',self)
        getattr(self, 'dispatcher.html')
        
    def MarkNotificationDispatch(self):
        """ set as dispatched """
        self.dispatched = 1
        
InitializeClass(IssueTrackerNotification)

