# IssueTrackerProduct
#
# Peter Bengtsson <mail@peterbe.com>
# License: ZPL
#


# python
import string
import poplib
from types import IntType, StringType, ListType, DictType

# Zope
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from OFS import SimpleItem, ObjectManager, Folder
from AccessControl import ClassSecurityInfo
from Globals import Persistent, InitializeClass, DTMLFile

# Product
import IssueTrackerUtils
import IssueTracker
from IssueTrackerConstants import *

class POP3Account(Folder.Folder, IssueTracker, Persistent):
    """ POP3Account class """
    
    meta_type = POP3ACCOUNT_METATYPE
    icon = '%s/issuetracker_pop3account.gif'%ICON_LOCATION

    _properties=({'id':'hostname',    'type': 'string', 'mode':'w'},
                 {'id':'username',    'type': 'string', 'mode':'w'},
                 )
    
    security=ClassSecurityInfo()
    
    def __init__(self, id, hostname, username, password):
        """ init """
        self.id = str(id)
        self.hostname = str(hostname)
        self.username = str(username)
        self._password = str(password)
        
    def getTitle(self):
        """ return hostname for title """
        return self.hostname
        
    def getId(self):
        """ return id """
        return self.id
    
    def getHostname(self):
        """ return hostname """
        return self.hostname
    
    def getUsername(self):
        """ return username """
        return self.username
    
    def getAcceptingEmails(self, ids=0):
        """ return accepting email objects here """
        if ids:
            return self.objectIds(ACCEPTINGEMAIL_METATYPE)
        else:
            return self.objectValues(ACCEPTINGEMAIL_METATYPE)
        
    def getAcceptingEmailbyTo(self, to, default=None):
        """ which accepting email here has 'to' 
        for email_address """
        to = to.lower().strip()
        for object in self.getAcceptingEmails():
            if object.email_address.lower().strip() == to:
                return object
        else:
            return default
        
    def editAccount(self, hostname=None, username=None,
                    password=None):
        """ change the attributes """
        if hostname is not None:
            self.hostname = hostname
        if username is not None:
            self.username = username
        if password is not None:
            self._password = password
            
    def createAcceptingEmail(self, id, email_address, defaultsections,
                             default_type, default_urgency,
                             send_confirm):
        """ create object and return it """
        acceptingemail = AcceptingEmail(id, email_address, defaultsections,
                                        default_type, default_urgency,
                                        send_confirm)
        self._setObject(id, acceptingemail)
        return getattr(self, id)
        
InitializeClass(POP3Account)



class AcceptingEmail(SimpleItem.SimpleItem,
                     IssueTracker):
    """ AcceptingEmail class """

    meta_type = ACCEPTINGEMAIL_METATYPE
    icon = '%s/issuetracker_acceptingemail.gif'%ICON_LOCATION
    
    meta_types = []

    _properties=({'id':'email_address',   'type': 'string', 'mode':'w'},
                 {'id':'defaultsections', 'type': 'lines',  'mode':'w'},
                 {'id':'default_type',    'type': 'string', 'mode':'w'},
                 {'id':'default_urgency', 'type': 'string', 'mode':'w'},
                 {'id':'send_confirm',    'type': 'boolean','mode':'w'},
                 )
    
    security=ClassSecurityInfo()

    #manage_options = (
        #{'label':'Properties', 'action':'manage_editNotifyableGroupForm'},
        #)
    
    def __init__(self, id, email_address, defaultsections,
                 default_type, default_urgency, send_confirm=1):
        """ init """
        self.id = str(id)
        self.email_address = str(email_address)
        if type(defaultsections) == type('s'):
            defaultsections = [defaultsections]
        self.defaultsections = defaultsections
        self.default_type = str(default_type)
        self.default_urgency = str(default_urgency)
        if send_confirm:
            self.send_confirm = true
        else:
            self.send_confirm = false

    def getId(self):
        """ return id """
        return self.id
    
    def getTitle(self):
        """ return email_address """
        return self.email_address
    
    def getEmailAddress(self):
        """ return email_address """
        return self.email_address
    
    def doSendConfirm(self):
        """ return send_confirm """
        return getattr(self, 'send_confirm', 0)
    
    def editDetails(self, email_address=None, defaultsections=None,
                    default_type=None, default_urgency=None,
                    send_confirm=None):
        """ edit details if not None """
        if email_address is not None:
            self.email_address = email_address
        if defaultsections is not None:
            if type(defaultsections) == type('s'):
                defaultsections = [defaultsections]
            self.defaultsections = defaultsections
        if default_type is not None:
            self.default_type = default_type
        if default_urgency is not None:
            self.default_urgency = default_urgency
        if send_confirm is not None:
            if send_confirm:
                self.send_confirm = true
            else:
                self.send_confirm = false
    
    def assertAllProperties(self):
        """ make sure accepting email has all properties """
        props = {'email_address':'', 'defaultsections':[], 
                 'default_type':self.default_type,
                 'default_urgency':self.default_urgency,
                 'send_confirm':1}
                 
        count = 0
        for key, default in props.items():
            if not self.__dict__.has_key(key):
                self.__dict__[key] = default
                count += 1
        return count    
    
InitializeClass(AcceptingEmail)