# Peter Bengtsson <mail@peterbe.com>
# Utility scripts for IssueTrackerProduct
#

# python 
import string
import sys
import re
import random
from random import shuffle
from types import StringType, ListType
    
# zope
from DateTime import DateTime
from DocumentTemplate import sequence
from Products.PythonScripts.standard import html_quote, newline_to_br, \
                     structured_text, url_quote

# Language constants
HOUR = 'hour'
HOURS = 'hours'
YEAR = 'year'
YEARS = 'years'
MONTH = 'month'
MONTHS = 'months'
DAY = 'day'
DAYS = 'days'
AND = 'and'

# entity characters
d={'<':"&lt;",  '>':"&gt;",      '':"&Agrave;",
'':"&acirc;",  '':"&Acirc;",   '':"&auml;",
'':"&Auml;",   '':"&aring;",   '':"&Aring;",
'':"&aelig;",  '':"&AElig;",   '':"&ccedil;",
'':"&Ccedil;", '':"&eacute;",  '':"&Eacute;",
'':"&egrave;", '':"&Egrave;",  '':"&ecirc;",
'':"&Ecirc;",  '':"&euml;",    '':"&Euml;",
'':"&iuml;",   '':"&Iuml;",    '':"&ocirc;",
'':"&Ocirc;",  '':"&ouml;",    '':"&Ouml;",
'':"&oslash;", '':"&Oslash;",  '':"&szlig;",
'':"&ugrave;", '':"&Ugrave;",  '':"&ucirc;",
'':"&Ucirc;",  '':"&uuml;",    '':"&Uuml;",
'':"&reg;",    '':"&copy;",    '':"&euro;",
"'":"&#39;",    '':"&quot;", '':"&quot;",
'':"&#39;", '':"&#quot;",   '':"&#quot;"}

def _minpositive(px,px1):
    if px<0:
        if px1>=0:
            return px1
        return -1
    if px1<0:
        return px
    return min(px,px1)

def addhrefs(text):
    tags=text.split("<")
    newtags=[]
    in_link=0
    for tag in tags:
        ltag=tag.lower()
        if not tag:
            newtags.append("")
            continue
        if ltag[0]=="a":
            in_link=1
        elif ltag[0:1]=="/a":
            in_link=0
        if not in_link:
            p=tag.find(">")
            if p<0:
                p=0
            tag_pre=tag[:p]
            tag=tag[p:]
            ltag=ltag[p:]
            while tag:
                is_maillink=0
                has_address=ltag.find("www.")
                has_address=_minpositive(has_address,ltag.find("http://"))
                has_address=_minpositive(has_address,ltag.find("ftp://"))
                has_address=_minpositive(has_address,ltag.find("@"))
                if has_address>=0:
                    address_start=has_address
                    while address_start>=p:
                        if tag[address_start] in \
                           (">"," ","\t","\r","\n",'"',"'"):
                            break
                        address_start-=1
                    if tag[address_start] in \
                       (">"," ","\t","\r","\n",'"',"'"):
                        address_start+=1
                    address_end=has_address
                    while address_end<len(tag):
                        if tag[address_end] in (" ","\t","\r","\n",'"',"'"):
                            break
                        address_end+=1
                    address=tag[address_start:address_end]
                    addlater=''
                    if address[-1] in [',','.',';',')']:
                        addlater=address[-1]
                        address=address[:-1]
                        
                    show = address
                    maxlength = 80
                    if len(show) > maxlength:
                        bit1 = show[:maxlength]
                        bit2 = show[maxlength:]
                        if len(bit2) > maxlength:
                            bit2 = "%s<br>%s"%(bit2[:maxlength], bit2[maxlength:])
                        show = "%s<br>%s"%(bit1, bit2)
                    if address.find("@")>=0 and not address.startswith('http'):
                        link='<a href="mailto:'+address+'">'+show+'</a>'
                    else:
                        if address[:3].lower()=="www":
                            link='<a href="http://'+address+'">'+show+'</a>'
                        else:
                            link='<a href="'+address+'">'+show+'</a>'
                    tag_pre+=tag[:address_start]+link+addlater
                    tag=tag[address_end:]
                    ltag=ltag[address_end:]
                else:
                    tag_pre+=tag
                    tag=ltag=""
            tag=tag_pre
        newtags.append(tag)
    return "<".join(newtags)


def safeId(id, nospaces=0):
    """ Just make sure it contains no dodgy characters """
    lowercase = 'abcdefghijklmnopqrstuvwxyz'
    digits = '0123456789'
    specials = '_-.'
    allowed = lowercase + lowercase.upper() + digits + specials
    if not nospaces:
        allowed = ' ' + allowed
    n_id=[]
    allowed_list = list(allowed)
    for letter in list(id):
        if letter in allowed_list:
            n_id.append(letter)
    return ''.join(n_id)


def ShowFilesize(bytes):
    """ Return nice representation of size """
    if bytes < 1024:
        return "1 Kb"
    elif bytes > 1048576:
        mb_bytes = '%0.02f'%(bytes / 1048576.0)
        return "%s Mb"%mb_bytes
    else:
        return "%s Kb"%int(bytes / 1024)
    

def preParseEmailString(es, sep=', ', names2emails={}, aslist=0):
    """ Take any string and strip out only a string
        of email addresses delimited by 'sep'
        "Peter <mail@peterbe.com>, John;peter@grenna.net, Joe"
        =>
        "mail@peterbe.com; peter@grenna.net"
        But suppose names2email={'joe','joey@host.com')
        then you would expect:
        "mail@peterbe.com; peter@grenna.net; joey@host.com"
    """
    real_emails=[]
    es = es.replace(';',' ').replace(',',' ')
    es = es.replace('>',' ').replace('<',' ')
    transtab = string.maketrans('/ ','  ')
    es = string.translate(es,transtab,'?&!()<=>*#[]{}')

    for e in es.split():
        if '@' in e and ValidEmailAddress(e) and e not in real_emails:
            real_emails.append(e)
        elif names2emails:
            for k,v in names2emails.items():
                if k != '':
                    e = e.lower().replace(k.lower(), v)
            if '@' in e and ValidEmailAddress(e) and e not in real_emails:
                real_emails.append(e)
    if real_emails:
        if aslist:
            return real_emails
        else:
            return sep.join(real_emails)
    else:
        return None
        
def getFromstring(name, email, email_html=1, email_must=0, email_pure=0):
    """
    First we filter which combs to care about;
    then we filter them.
    """
    name_html_quoted =html_quote(name)
    email_html_quoted=html_quote(email)
    
    passit = 0
       
    # first filter
    if name <> '' or email <> '':
        if email_must:
            if ValidEmailAddress(email):
                passit = 1
            else:
                passit = 0
        else:
            passit = 1
    
    # then display
    if not passit:
        return 0
    else:
        if email_pure and ValidEmailAddress(email):
            return email
        elif email == '':
            return name
        else:
            if email_html:
                return string.strip('%s &lt;<a href="mailto:%s">%s</a>&gt;'%\
                                    (name_html_quoted, email_html_quoted,
                                     email_html_quoted))
            else:
                return string.strip('%s <%s>'%(name, email))
    return 0

def AddParam2URL(url, params={}):
    """ return url and append params but be aware of existing params """
    p='?'
    if p in url:
        p = '&'
    url = url + p
    for key, value in params.items():
        url = url + '%s=%s&'%(key, url_quote(value))
    return url[:-1]

def getLaterDate(days=365):
    """ return a dateobject in future in rfc822 format """
    later = getNowdate()+days
    return later.rfc822()

def fixDictofLists(dict):
    " throw it a dictionary and it returns the values lowercased "
    for key,value in dict.items():
        if type(value) == ListType:
            dict[key] = _lowercaseList(value)
        elif type(value) == StringType:
            dict[key] = value.lower()
    return dict

def _lowercaseList(lst):
    " lowercase and strip all items in a list "
    nlst=[]
    for l in lst:
        nlst.append(l.strip().lower())
    return nlst

def getRandomString(length=10, loweronly=1, numbersonly=0):
    """ return a very random string """
    if numbersonly:
        l = list('0123456789')
    else:
        lowercase = 'abcdefghijklmnopqrstuvwxyz'+'0123456789'
        l = list(lowercase + lowercase.upper())
    shuffle(l)
    s = string.join(l,'')
    if len(s) < length:
        s = s + getRandomString(loweronly=1)
    s = s[:length]
    if loweronly:
        return s.lower()
    else:
        return s
    
    

def timeSince(firstdate, seconddate, afterword=None):
    """
     Use two date objects to return in plain english the difference between them.
     E.g. "3 years and 2 days"
      or  "1 year and 3 months and 1 day"
     If less than 1 day, return number of hours.

     If there is "no difference" between them, return false.
    """

    fdo = firstdate
    sdo = seconddate

    day_difference = int(abs(sdo-fdo))

    years = day_difference/365
    months = (day_difference % 365)/30
    days = (day_difference % 365) % 30

    if days == 0 and months == 0:
        # use hours
        hours=int(abs(sdo-fdo)*24)
        if hours == 1:
            if afterword is None:
                return "1 %s" % (HOUR)
            else:
                return "1 %s %s" % (HOUR, afterword)
        elif hours > 0:
            if afterword is None:
                return "%s %s" % (hours, HOURS)
            else:
                return "%s %s %s" % (hours, HOURS, afterword)
        else:
            # if the difference is smaller than 1 hour, 
            # return it false
            return 0
    else:
        s = []
        if years == 1:
            s.append('1 %s'%(YEAR))
        elif years > 1:
            s.append('%s %s'%(years,YEARS))
        
        if months == 1:
            s.append('1 %s'%MONTH)
        elif months > 1:
            s.append('%s %s'%(months,MONTHS))
        
        if days == 1:
            s.append('1 %s'%DAY)
        elif days > 1:
            s.append('%s %s'%(days,DAYS))
        
        if len(s)>1:
            if afterword is None:
                return "%s" % (string.join(s,' %s '%AND))
            else:
                return "%s %s" % (string.join(s,' %s '%AND), afterword)
        else:
            if afterword is None:
                return "%s" % (s[0])
            else:
                return "%s %s" % (s[0], afterword)






def DateStr2DateDate(str):
    """ convert a date as a string to a proper date """
    return DateTime(str)

def cookIdAndTitle(s):
    """ if s='image1~Image One.gif' then return ['image1.gif','Image One']

        Testwork:
        s='image1~Image One.gif'    => ['image1.gif','Image One']
        s='image1.gif'              => ['image1.gif','']
        s='image1~.gif'             => ['image1.gif','']
        s='image1~Image One.gif.gif'=> ['image1.gif','Image One.gif']
    """
    if s.find('~') == -1:
        return s, ''
    splitted = s.split('~',1)
    id = splitted[0]
    rest = s.replace(id+'~','')
    if rest.rfind('.') == -1:
        return id, rest
    else:
        pos_last_dot = rest.rfind('.')
        ext = rest[pos_last_dot:]
        id = id.strip() + ext.strip()
        title = rest[0:pos_last_dot].strip()
        return id, title

def LineIndent(text, indent):
    """ indent each new line with 'indent' """
    text = indent+text
    text = text.replace('\n','\n%s'%indent)
    return text


def ShowDescription(text, display_format=''):
    """
    Display text, using harmless HTML
    """

    if display_format == 'structuredtext':
        #st=_replace_special_chars(text)
        st=text

        for k,v in {'<':'&lt;', '>':'&gt;',
                    '[':'|[|'}.items():
            st = st.replace(k,v)

        st = structured_text(st)

        for k,v in {'&amp;lt;':'&lt;', '&amp;gt;':'&gt;',
                    '|[|':'['}.items():
            st = st.replace(k,v)
            
        # BUG in structured_text in Zope 2.4.0
        # it appends these annoying tags.
        for tag in ['<html>','<body>','</body>','</html>']:
            st = st.replace(tag, '')
        
        #return _replace_special_chars(st, 0)
        st = addhrefs(st)
        return st
    elif display_format == 'html':
        return text
    else:
        t = html_quote(text)
        t = t.replace('&amp;lt;','&lt;').replace('&amp;gt;','&gt;')
        t = addhrefs(t)
        t = newline_to_br(t)
        return t

def _replace_special_chars(text, simplify=1, html_encoding=0):
    """ Replace special characters with placeholder keys back and forth.
        The reason for doing this is that structured_text() doesn't support
        special characters such as .
    """
    reps = {'Aring':'',  'Auml':'', 'Ouml':'',
            'aring':'',  'auml':'', 'ouml':'',
            'Egrave':'', 'Eacute':'',
            'egrave':'', 'eacute':'',
            'Uuml':'Ü',   'uuml':'ü'}

    if simplify:
        for k, v in reps.items():
            if html_encoding:
                k='&%s;'%k
            else:
                k='__%s__'%k
            text = text.replace(v,k)
    else:
        for k, v in reps.items():
            k='__%s__'%k
            text = text.replace(k,v)
    return text
    

def getNowdate():
    """
    This method determines the format 
    with which new objects get a date property
    """
    return DateTime()


def _ShouldBeNone(result): return result is not None
def _ShouldNotBeNone(result): return result is None

tests = (
    # Thank you Bruce Eckels for these
  (re.compile("^[0-9a-zA-Z\.\-\_]+\@[0-9a-zA-Z\.\-]+$"), _ShouldNotBeNone, "Failed a"),
  (re.compile("^[^0-9a-zA-Z]|[^0-9a-zA-Z]$"), _ShouldBeNone, "Failed b"),
  (re.compile("([0-9a-zA-Z]{1})\@."), _ShouldNotBeNone, "Failed c"),
  (re.compile(".\@([0-9a-zA-Z]{1})"), _ShouldNotBeNone, "Failed d"),
  (re.compile(".\.\-.|.\-\..|.\.\..|.\-\-."), _ShouldBeNone, "Failed e"),
  (re.compile(".\.\_.|.\-\_.|.\_\..|.\_\-.|.\_\_."), _ShouldBeNone, "Failed f"),
  (re.compile(".\.([a-zA-Z]{2,3})$|.\.([a-zA-Z]{2,4})$"), _ShouldNotBeNone, "Failed g"),
)
def ValidEmailAddress(address, debug=None):
    for test in tests:
        if test[1](test[0].search(address)):
            if debug: return test[2]
            return 0
    return 1

def ss(s):
    """ simple string """
    return s.strip().lower()

  
def test():

    print "TEST getLaterDate()"
    print getLaterDate()
    print getLaterDate(14)

    print "----------\n"
    print "TEST AddParam2URL()"
    print AddParam2URL('http://www.peterbe.com')
    print AddParam2URL('http://www.peterbe.com',{'a':'A','b':'B'})
    message="""Line1
Line2"""
    print AddParam2URL('http://www.peterbe.com?o=O',{'a':message,'b':'B spaced'})

    print "----------\n"
    print "TEST preParseEmailString()"
    print preParseEmailString("mail@peterbe.com;JOE,peter@grenna.net,Peter Bengtsson<ppp@ppp.se>; Peter <pete@peterbe.com", \
                              names2emails={'Joe':'joey@peterbe.com', 'Peter bengtsson':'ppp@ppp.se'})
    print preParseEmailString("mail@ ", \
                              names2emails={'Joe':'joey@peterbe.com', 'Peter bengtsson':'ppp@ppp.se'})
    

    print " ++ cookIdAndTitle() ++ "
    print cookIdAndTitle('image1~Image One.gif'), " AND ['image1.gif','Image One']"
    print cookIdAndTitle('image1.gif'), " AND ['image1.gif','']"
    print cookIdAndTitle('image1~.gif'), " AND ['image1.gif','']"
    print cookIdAndTitle('image1~Image One.gif.gif'), " AND ['image1.gif','Image One.gif']"
    print cookIdAndTitle('monk ~ Monkey.dtml'), " AND ['monk.dtml','Monkey']"

    print " ++ ValidEmailAddress() ++ "
    print ValidEmailAddress(''), " AND 0"
    print ValidEmailAddress('issuetracker@peterbe.com'), " AND 1"
    print ValidEmailAddress('issuetracker @peterbe.com'), " AND 0"

    print " ++ getRandomString() ++ "
    print getRandomString(), " "
    print getRandomString(length=50), " AND (length=50)"
    print getRandomString(loweronly=1), " AND (loweronly=1)"
    
if __name__=='__main__':
    test()
